"""Event handlers for CMFNotification.

See ``events.txt`` for further details.

$Id$
"""

from Products.CMFCore.utils import getToolByName

from Products.CMFNotification.NotificationTool import ID

import transaction


def onObjectInitializedEvent(obj, event):
    """Subscriber for ``ObjectInitializeEvent``."""
    ntool = getToolByName(obj, ID, None)
    if ntool is not None:
        ntool.onItemCreation(obj)


def onObjectClonedEvent(obj, event):
    """Subscriber for ``ObjectClonedEvent``."""
    ntool = getToolByName(obj, ID, None)
    if ntool is not None:
        ntool.onItemCreation(obj)


def onObjectEditedEvent(obj, event):
    """Subscriber for ``ObjectModifiedEvent``."""
    ntool = getToolByName(obj, ID, None)
    if ntool is not None:
        ntool.onItemModification(obj)


def onItemRemoval_hook(status, *args, **kwargs):
    """ transaction hook to call item removal handler"""
    if status:
        obj = args[0]
        ntool = getToolByName(obj, ID, None)
        if ntool is not None:
            ntool.onItemRemoval(*args, **kwargs)


def onObjectRemovedEvent(obj, event):
    """Subscriber for ``ObjectRemovedEvent``."""
    # if we're on delete_confirmation, only trigger if submitted and not cancelled
    if obj.REQUEST.getURL().endswith('delete_confirmation') and \
       (obj.REQUEST.form.get('form.button.Cancel', None) == 'Cancel' or
                    not obj.REQUEST.form.get('form.submitted')):
        return
    currentTransaction = transaction.get()
    # check if we already set the hook
    if len([x for x in currentTransaction.getAfterCommitHooks() 
              if x[0] == onItemRemoval_hook and obj in x[1]]) == 0:
        # we wait for transaction completion as removal can be aborted
        currentTransaction.addAfterCommitHook(onItemRemoval_hook, args=(obj, ))


def onActionSucceededEvent(obj, event):
    """Subscriber for ``ActionSucceededEvent``."""
    ntool = getToolByName(obj, ID, None)
    if ntool is not None:
        ntool.onWorkflowTransition(obj, event.action)


def onDiscussionItemAddedEvent(obj, event):
    """Subscriber for ``ObjectAddedEvent`` on discussion items."""
    ntool = getToolByName(obj, ID, None)
    if ntool is not None:
        ntool.onDiscussionItemCreation(obj)



"""Install method.

$Id$
"""

from zope.component import getUtility
from zope.component import getMultiAdapter

from Products.CMFCore.utils import getToolByName

from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletAssignmentMapping

from Products.CMFNotification.config import LAYER_NAME
from Products.CMFNotification.config import PORTLET_NAME
from Products.CMFNotification.config import PROJECT_NAME
from Products.CMFNotification.NotificationTool import ID as TOOL_ID


def uninstall(context):
    """Uninstall CMFNotification."""
    portal = getToolByName(context, 'portal_url').getPortalObject()
    ps = getToolByName(portal, "portal_setup")
    irs = ps.getImportStepRegistry()

    ## Import GenericSetup uninstallation profile
    setup_tool = getToolByName(context, 'portal_setup')
    setup_tool.runAllImportStepsFromProfile('profile-Products.CMFNotification:uninstall')

    ## Remove portlet
    for name in ('plone.rightcolumn', 'plone.leftcolumn'):
        manager = getUtility(IPortletManager,
                             name=name,
                             context=portal)
        column = getMultiAdapter((portal, manager),
                                 IPortletAssignmentMapping,
                                 context=portal)
        if PORTLET_NAME in column:
            del column[PORTLET_NAME]

    ## Remove the import and export steps from portal_setup
    ## Since 'remove=True' doesn't seem to work for import/export
    ## steps, here we manually remove import/export steps (Kurt)
    if 'export_cmfnotification' in ps.listExportSteps():
        ps.manage_deleteExportSteps(['export_cmfnotification',])
    if 'import_cmfnotification' in irs.listSteps():
        ps.manage_deleteImportSteps(['import_cmfnotification',])

    ## Remove installation step for the subscription portlet
    if u'' in irs.listSteps():
        ## The 2.2-dev step didn't specify an ID. So, check for a step
        ## called '' (Kurt)
        irs.unregisterStep('')
    if 'import_cmfnotification_portlet' in irs.listSteps():
        irs.unregisterStep('import_cmfnotification_portlet')

    ## Remove configlet
    panel = getToolByName(portal, 'portal_controlpanel')
    if panel is not None:
        panel.unregisterConfiglet('cmfnotification_configuration')
        panel.unregisterConfiglet('cmf_notification_unsubscribemenu')

    ## For some reason, under Plone 3, the tool is not removed by the
    ## GenericSetup 'uninstall' profile.
    tool = getToolByName(portal, TOOL_ID, None)
    if tool is not None:
        portal.manage_delObjects([TOOL_ID])

    return '%s has been successfully uninstalled.' % PROJECT_NAME



"""CMFNotification Generic Setup handlers.

$Id$
"""

from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter

from Products.GenericSetup.interfaces import IBody

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.permissions import setDefaultRoles

from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletAssignmentMapping

from Products.CMFNotification.browser import portlet
from Products.CMFNotification.config import PORTLET_NAME
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION


_FILENAME = 'cmfnotification.xml'


def importCMFNotification(context):
    """Import CMFNotification tool properties."""
    site = context.getSite()
    logger = context.getLogger('CMFNotification')
    ntool = getToolByName(site, 'portal_notification')

    body = context.readDataFile(_FILENAME)
    if body is None:
        logger.info('Nothing to import.')
        return

    importer = queryMultiAdapter((ntool, context), IBody)
    if importer is None:
        logger.warning('Import adapter missing.')
        return

    importer.body = body
    logger.info('CMFNotification tool imported.')


def exportCMFNotification(context):
    """Export CMFNotification tool properties."""
    site = context.getSite()
    logger = context.getLogger('CMFNotification')
    ntool = getToolByName(site, 'portal_notification', None)
    if ntool is None:
        logger.info('Nothing to export.')
        return

    exporter = queryMultiAdapter((ntool, context), IBody)
    if exporter is None:
        logger.warning('CMFNotification adapter missing.')
        return

    context.writeDataFile(_FILENAME, exporter.body, exporter.mime_type)
    logger.info('MemberData tool exported.')


def addPortlet(context):
    """Add CMFNotification portlet to the right column."""
    portal = context.getSite()
    rightColumn = getUtility(IPortletManager,
                             name=u'plone.rightcolumn',
                             context=portal)
    right = getMultiAdapter((portal, rightColumn),
                            IPortletAssignmentMapping,
                            context=portal)
    if PORTLET_NAME not in right:
        right[PORTLET_NAME] = portlet.Assignment()


def addPermissions(context):
    """Add specific permissions."""
    setDefaultRoles(SUBSCRIBE_PERMISSION, ('Manager', ))



"""Tests CMFNotification event handlers.

$Id$
"""

import transaction
from zope import event

from Products.CMFCore.utils import getToolByName

from Products.Archetypes.event import ObjectInitializedEvent

from Products.CMFNotification.NotificationTool import ID as NTOOL_ID
from Products.CMFNotification.events.handlers import onItemRemoval_hook

from Products.CMFNotification.tests.base import FakeMailHost
from Products.CMFNotification.tests.base import CMFNotificationTestCase

## FIXME: would be nice to check log warnings, too.


class TestNotification(CMFNotificationTestCase):
    """Make sure that the notification works, which would be just
    great, since this is the very purpose of this product.

    The current tests are a bit verbose and very similar. They could
    be generated dynamically, but I think debugging would be much
    harder.

    Note that these tests do not check the contents of the e-mails. It
    only checks the number of e-mails that have been sent.
    """

    def failUnlessSent(self, n):
        """Make the test fail unless 'n' mails have been sent."""
        mh = self.portal.MailHost
        self.failUnless(len(mh.getSentList()) == n,
                        "Expected %s mail(s), but "\
                        "sent %s" % (n, len(mh.getSentList())))


    def afterSetUp(self):
        """Called before each tests.

        This method:

        - replaces the mailhost by a fake one which does not send
        mails but keeps them in a list;

        - create dummy users and content.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        ntool.manage_changeProperties(item_creation_notification_enabled=False)
        self.createTestUsersAndContent()
        ntool.manage_changeProperties(item_creation_notification_enabled=True)

        ## Set up a fake mail host
        if 'MailHost' in portal.objectIds():
            portal._delObject('MailHost')
        portal._setObject('MailHost', FakeMailHost())


    def testOnItemCreation(self):
        """Test notification on item creation."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        mh = portal.MailHost
        self.login('manager')

        ## Set correct rules so that 3 mails should be sent.
        changeProperty('item_creation_notification_enabled', True)
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:creation_mail_notification'])

        portal.invokeFactory('Document', 'document')
        ## See 'events/events.txt' for futher details about this
        ## manually fired event.
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(1)
        portal.manage_delObjects(['document'])
        mh.clearSentList()

        ## Set workflow initial state to 'publish', thus showing the
        ## new item to every users.
        wtool.simple_publication_workflow.initial_state = 'published'
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(3)
        portal.manage_delObjects(['document'])
        mh.clearSentList()

        ## Disable notification
        changeProperty('item_creation_notification_enabled', False)
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(0)
        portal.manage_delObjects(['document'])
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('item_creation_notification_enabled', True)
        ntool.manage_changeProperties(on_item_creation_users='* :: python: []')
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(0)
        portal.manage_delObjects(['document'])
        mh.clearSentList()

        ## Set the notified users list to "everybody" but ask for a
        ## missing mail template
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:does_not_exist'])
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(0)
        portal.manage_delObjects(['document'])
        mh.clearSentList()


    def testOnItemCopiedAndPasted(self):
        """Test notification when an item is copied and then
        pasted.

        This test is actually a copy of ``testOnItemCreation()``,
        since CMFNotification shoudl have the same behaviour in both
        events.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        mh = portal.MailHost
        self.login('manager')

        def copyAndPaste():
            cb = portal.folder.manage_copyObjects(['document1'])
            portal.manage_pasteObjects(cb)

        ## Set correct rules so that 3 mails should be sent.
        changeProperty('item_creation_notification_enabled', True)
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:creation_mail_notification'])
        copyAndPaste()
        self.failUnlessSent(1)
        portal.manage_delObjects(['document1'])
        mh.clearSentList()

        ## Set workflow initial state to 'publish', thus showing the
        ## new item to every users.
        wtool.simple_publication_workflow.initial_state = 'published'
        copyAndPaste()
        self.failUnlessSent(3)
        portal.manage_delObjects(['document1'])
        mh.clearSentList()

        ## Disable notification
        changeProperty('item_creation_notification_enabled', False)
        copyAndPaste()
        self.failUnlessSent(0)
        portal.manage_delObjects(['document1'])
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('item_creation_notification_enabled', True)
        ntool.manage_changeProperties(on_item_creation_users='* :: python: []')
        copyAndPaste()
        self.failUnlessSent(0)
        portal.manage_delObjects(['document1'])
        mh.clearSentList()

        ## Set the notified users list to "everybody" but ask for a
        ## missing mail template
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:does_not_exist'])
        copyAndPaste()
        self.failUnlessSent(0)
        portal.manage_delObjects(['document1'])
        mh.clearSentList()


    def testOnItemCutAndPasted(self):
        """Test notification when an item is cut and then pasted."""
        ## Since there is currently no special rule for cut and paste,
        ## we should not send anything.
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')

        ## Enable some rules just to make sure that none of them
        ## match.
        changeProperty('item_creation_notification_enabled', True)
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:creation_mail_notification'])
        changeProperty('item_modification_notification_enabled', True)
        changeProperty('on_item_modification_users', ['* :: *'])
        changeProperty('on_item_modification_mail_template',
                       ['* :: string:modification_mail_notification'])
        changeProperty('item_wf_transition_notification_enabled', True)
        changeProperty('on_wf_transition_modification_users', ['* :: *'])
        changeProperty('on_wf_transition_modification_mail_template',
                       ['* :: string:workflow_mail_notification'])

        transaction.savepoint() ## We need this to cut/paste objects.
        cb = portal.folder.manage_cutObjects(['document1'])
        portal.manage_pasteObjects(cb)
        self.failUnlessSent(0)


    def testOnItemRenamed(self):
        """Test notification when an item is renamed."""
        ## Since there is currently no special rule for this event, we
        ## should not send anything.
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')

        ## Enable some rules just to make sure that none of them
        ## match.
        changeProperty('item_creation_notification_enabled', True)
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:creation_mail_notification'])
        changeProperty('item_modification_notification_enabled', True)
        changeProperty('on_item_modification_users', ['* :: *'])
        changeProperty('on_item_modification_mail_template',
                       ['* :: string:modification_mail_notification'])
        changeProperty('item_wf_transition_notification_enabled', True)
        changeProperty('on_wf_transition_modification_users', ['* :: *'])
        changeProperty('on_wf_transition_modification_mail_template',
                       ['* :: string:workflow_mail_notification'])

        transaction.savepoint() ## We need this to cut/paste objects.
        portal.folder.manage_renameObjects(('document1', ),
                                           ('renamed-document', ))
        self.failUnlessSent(0)


    def testOnItemModification(self):
        """Test notification on item modification."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        document.unmarkCreationFlag()
        mh = portal.MailHost
        self.login('manager')

        ## Set correct rules so that an e-mail should be sent.
        changeProperty('item_modification_notification_enabled', True)
        changeProperty('on_item_modification_users', ['* :: *'])
        changeProperty('on_item_modification_mail_template',
                       ['* :: string:modification_mail_notification'])

        ## Calling the 'edit()' method does not trigger notification.
        document.edit(title='New title')
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Calling a field mutator does not trigger notification.
        document.setTitle('New title')
        self.failUnlessSent(0)
        mh.clearSentList()

        ## However, editing the item through the edit form does
        ## trigger notification.
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Publish document so that other users can view it (and thus
        ## receive notification).
        wtool.doActionFor(document, 'publish')
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Disable notification
        changeProperty('item_modification_notification_enabled', False)
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('item_modification_notification_enabled', True)
        changeProperty('on_item_modification_users',
                       ['* :: python: []'])
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        changeProperty('on_item_modification_users', ['* :: *'])
        changeProperty('on_item_modification_mail_template',
                       ['* :: string:does_not_exist'])
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(0)
        mh.clearSentList()


    def testOnItemRemoval(self):
        """Test notification on item removal."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        mh = portal.MailHost
        self.login('manager')

        ## Set correct rules so that 3 mails should be sent.
        changeProperty('item_removal_notification_enabled', True)
        changeProperty('on_item_removal_users', ['* :: *'])
        changeProperty('on_item_removal_mail_template',
                       ['* :: string:removal_mail_notification'])

        portal.invokeFactory('Document', 'document')
        ## See 'events/events.txt' for futher details about this
        ## manually fired event.
        event.notify(ObjectInitializedEvent(portal['document']))
        obj = portal['document']
        # Do removal
        portal.manage_delObjects(['document'])
        onItemRemoval_hook(True, obj) # Normally this gets called by an after 
        # commit hook. But using transcation.commit() here to trigger it causes
        # all kinds of problems.
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Set workflow initial state to 'publish', thus showing the
        ## new item to every users.
        wtool.simple_publication_workflow.initial_state = 'published'
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        obj = portal['document']
        portal.manage_delObjects(['document'])
        onItemRemoval_hook(True, obj) 
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Disable notification
        changeProperty('item_removal_notification_enabled', False)
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        portal.manage_delObjects(['document'])
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('item_removal_notification_enabled', True)
        ntool.manage_changeProperties(on_item_removal_users='* :: python: []')
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        portal.manage_delObjects(['document'])
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the notified users list to "everybody" but ask for a
        ## missing mail template
        changeProperty('on_item_removal_users', ['* :: *'])
        changeProperty('on_item_removal_mail_template',
                       ['* :: string:does_not_exist'])
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        portal.manage_delObjects(['document'])
        self.failUnlessSent(0)
        mh.clearSentList()


    def testOnWorkflowTransition(self):
        """Test notification on workflow transition."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        mh = portal.MailHost
        self.login('manager')

        ## Set correct rules so that 3 mails should be sent.
        changeProperty('wf_transition_notification_enabled', True)
        changeProperty('on_wf_transition_users',
                       ['* :: *'])
        changeProperty('on_wf_transition_mail_template',
                       ['* :: string:workflow_mail_notification'])
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Retract 'document' and thus hide it from everybody except
        ## manager
        wtool.doActionFor(document, 'retract')
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Disable notification
        changeProperty('wf_transition_notification_enabled', False)
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('wf_transition_notification_enabled', True)
        changeProperty('on_wf_transition_users',
                       ['* :: python: []'])
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        changeProperty('on_wf_transition_users', ['* :: *'])
        changeProperty('on_wf_transition_mail_template',
                       ['* :: string:does_not_exist'])
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(0)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()


    def testOnMemberRegistration(self):
        """Test notification on member registration."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        rtool = getToolByName(portal, 'portal_registration')
        mtool = getToolByName(portal, 'portal_membership')
        deleteMember = lambda userid: mtool.deleteMembers([userid, ])
        mh = portal.MailHost
        userid = 'a_new_member'
        self.login('manager')

        ## Set correct rules so that 3 mails should be sent.
        changeProperty('member_registration_notification_enabled', True)
        changeProperty('on_member_registration_users', ['* :: *'])
        changeProperty('on_member_registration_mail_template',
                       ['* :: string:registration_mail_notification'])
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(3)
        deleteMember(userid)
        mh.clearSentList()

        ## Disable notification
        changeProperty('member_registration_notification_enabled', False)
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('member_registration_notification_enabled', True)
        changeProperty('on_member_registration_users',
                       ['* :: python: []'])
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        changeProperty('on_member_registration_users', ['* :: *'])
        changeProperty('on_member_registration_mail_template',
                       ['* :: string:does_not_exist'])
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(0)
        deleteMember(userid)
        mh.clearSentList()


    def testOnMemberModification(self):
        """Test notification on member modification."""
        pass ## FIXME: implement me! (Damien)


    def testOnDiscussionItemCreation(self):
        """Test notification on discussion item creation."""
        return ## FIXME: test deactivated until feature is implemented
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        wtool = getToolByName(portal, 'portal_workflow')
        document = portal.folder.document1
        document.allow_discussion = True
        mh = portal.MailHost
        addDiscussionItem = lambda obj: obj.discussion_reply('Subject',
                                                             'Body text')
        self.login('manager')

        ## Set correct rules so that 1 e-mail should be sent.
        changeProperty('on_discussion_item_creation_notification_enabled',
                       True)
        changeProperty('on_discussion_item_creation_users',
                       ['* :: *'])
        changeProperty('on_discussion_item_creation_mail_template',
                       ['* :: string:discussion_mail_notification'])
        addDiscussionItem(document)
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Publish 'document' so that everybody can view it (and thus
        ## receive a notification)
        wtool.doActionFor(document, 'publish')
        addDiscussionItem(document)
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Reply to a discussion item
        discussion_item = document.talkback.objectValues()[0]
        discussion_item = discussion_item.__of__(document.talkback)
        addDiscussionItem(discussion_item)
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Disable notification
        changeProperty('on_discussion_item_creation_notification_enabled',
                       False)
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Enable notification but set the notified users list to []
        changeProperty('discussion_item_creation_notification_enabled',
                       True)
        changeProperty('on_discussion_item_creation_users',
                       ['* :: python: []'])
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()

        ## Set the notified users list to <everybody> but ask for a
        ## missing mail template
        changeProperty('on_discussion_item_creation_users', ['* :: *'])
        changeProperty('on_discussion_item_creation_mail_template=',
                       ['* :: string:does_not_exist'])
        addDiscussionItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()


    def testHandlersDoNotOverlap(self):
        """Test that handlers do not overlap themselves, i.e. that at
        most one handler is called for any event.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        wtool = getToolByName(portal, 'portal_workflow')
        rtool = getToolByName(portal, 'portal_registration')
        mtool = getToolByName(portal, 'portal_membership')
        mh = portal.MailHost
        document = portal.folder.document1

        self.login('manager')

        ## Enable everything
        users = ['* :: *']
        properties = {
           'item_creation_notification_enabled': True,
           'on_item_creation_users': users,
           'on_item_creation_mail_template': ['* :: string:creation_mail_notification'],
           'item_modification_notification_enabled': True,
           'on_item_modification_users': users,
           'on_item_modification_mail_template': ['* :: string:modification_mail_notification'],
           'wf_transition_notification_enabled': True,
           'on_wf_transition_users': users,
           'on_wf_transition_mail_template': ['* :: string:workflow_mail_notification'],
           'member_registration_notification_enabled': True,
           'on_member_registration_users': users,
           'on_member_registration_mail_template':['* :: string:registration_mail_notification'],
           'discussion_item_creation_notification_enabled': True,
           'on_discussion_item_creation_users': users,
           'on_discussion_item_creation_mail_template': ['* :: string:discussion_mail_notification'],
           }
        ntool.manage_changeProperties(properties)

        ## Test item creation
        portal.invokeFactory('Document', 'document')
        event.notify(ObjectInitializedEvent(portal['document']))
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Test item modification
        document.processForm(data=1, values={'title': 'New title'})
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Test workflow transition
        wtool.doActionFor(document, 'publish')
        self.failUnlessSent(3)
        wtool.doActionFor(document, 'retract')
        mh.clearSentList()

        ## Test member registration
        userid = 'a_new_member'
        rtool.addMember(userid, 'password', properties=None)
        self.failUnlessSent(3)
        mtool.deleteMembers([userid, ])
        mh.clearSentList()

        ## FIXME: Test member modification

        ## Test discussion item creation
        ## FIXME: uncomment when implemented
        #document.allow_discussion = True
        #document.discussion_reply('Subject', 'Body text')
        #self.failUnlessSent(1)
        #mh.clearSentList()


    def testLabels(self):
        """Test template selection with labels."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        modifyItem = lambda obj: obj.processForm(data=1,
                                                 values={'title': 'New title'})
        wtool = getToolByName(portal, 'portal_workflow')
        mh = portal.MailHost
        document = portal.folder.document1

        ## We will only test labels on item modification.
        self.login('manager')
        wtool.doActionFor(document, 'publish')
        changeProperty('item_modification_notification_enabled', True)

        ## A simple configuration with two labels.
        users = ['* :: python: ["manager"] :: one',
                 '* :: python: ["manager", "member1", "member2"] :: two']
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "two" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(4) # 2 for manager; 1 for member1; 1 for member2
        mh.clearSentList()

        ## A special case, where the first rule match all users, which
        ## disable next rules. I (Damien) am not particularly
        ## convinced that this is really the behaviour than a user
        ## would expect, but I do not have any strong argument against
        ## it, neither. :)
        users = ['* :: * :: one',
                 '* :: python: ["manager", "member1"] :: two']
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "two" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(3) # 1 for each user
        mh.clearSentList()

        ## A configuration with a non-existent rule.
        users = ['* :: python: ["manager"] :: one',
                 '* :: python: ["manager", "member1"] :: two']
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "non-existente" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(1) # 1 for manager
        mh.clearSentList()

        ## A configuration with an error in a rule.
        users = ['* :: python: ["manager"] :: one',
                 '* :: python: ["manager", "member1"] :: two']
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "two" :: string:non_existent']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(1) # 1 for manager
        mh.clearSentList()

        ## A configuration with a label that is used twice.
        users = ['* :: python: ["manager"] :: one',
                 '* :: python: ["manager", "member1"] :: one']
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "two" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(3) # 2 for manager; 1 for member1
        mh.clearSentList()

        ## Test that manually subscribed users are available under the
        ## empty label.
        changeProperty('extra_subscriptions_enabled', True)
        ntool._subscriptions[ntool._getPath(document)] = {'member1': ('mail',)}
        users = []
        templates = ['python: label == "one" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(0)
        mh.clearSentList()
        templates = ['python: label == "one" :: string:modification_mail_notification',
                     'python: label == "" :: string:modification_mail_notification']
        changeProperty('on_item_modification_users', users)
        changeProperty('on_item_modification_mail_template', templates)
        modifyItem(document)
        self.failUnlessSent(1)
        mh.clearSentList()


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestNotification))
    return suite



"""Utility methods for CMFNotification.

$Id$
"""

import re
from email.Header import Header

from AccessControl import Unauthorized
from Products.PageTemplates.Expressions import getEngine
from Products.PageTemplates.Expressions import SecureModuleImporter

from Products.CMFCore.utils import getToolByName
from Products.CMFCore.WorkflowCore import WorkflowException


EMAIL_ADDRESS_IN_HEADER_REGEXP = re.compile('(?:\s*,\s*)?(.*?) <(.*?)>')


def getBasicBindings(obj):
    """Return basic bindings to build expression context and provide
    options to ZPT mail templates.
    """
    mtool = getToolByName(obj, 'portal_membership')
    wtool = getToolByName(obj, 'portal_workflow')
    try:
        current_state = wtool.getInfoFor(obj, 'review_state')
    except WorkflowException:
        current_state = None

    return {
        'current_state': current_state,
        'author': mtool.getAuthenticatedMember(),
        }


def getExpressionContext(obj, extra_bindings=None):
    """An expression context provides names for TALES expressions.

    Based on ``DCWorkflow.Expression`` and ``CMFCore.Expression``. We
    add ``getBasicBindings()`` and then ``extra_bindings`` if is it
    provided.
    """
    portal = getToolByName(obj, 'portal_url').getPortalObject()

    data = {
        'here': obj,
        'context': obj,
        'nothing': None,
        'portal': portal,
        'request': getattr(obj, 'REQUEST', None),
        'modules': SecureModuleImporter,
        }
    data.update(getBasicBindings(obj))

    if extra_bindings is not None:
        data.update(extra_bindings)

    return getEngine().getContext(data)


def encodeMailHeaders(message, encoding):
    """Return ``message`` with correctly encoded headers.

    The following headers are encoded: ``From``, ``Reply-to``,
    ``Sender``, ``Cc`` and ``Subject``.
    """
    mout = []
    lines = message.split('\n')
    for line_i in range(0, len(lines)):
        line = lines[line_i]
        if not line:
            break ## End of headers block.
        header = line[:line.find(':')]
        if header.lower() in ('from', 'reply-to', 'sender',
                              'cc', 'subject'):
            value = line[len(header) + 1:].lstrip()
            if header.lower() in ('from', 'reply-to', 'sender', 'cc'):
                ## We must not encode e-mail addresses.
                addresses = EMAIL_ADDRESS_IN_HEADER_REGEXP.findall(value)
                if addresses:
                    addresses = [(Header(s, encoding).encode(), addr) \
                                 for s, addr in addresses]
                    value = ', '.join(['%s <%s>' % (s, addr) \
                                       for (s, addr) in addresses])
            else:
                value = Header(value, encoding).encode()
            mout.append('%s: %s' % (header, value))
            continue
        mout.append(line)
    mout.extend(lines[line_i:])
    return '\n'.join(mout)


def getPreviousWorkflowState(obj):
    """Return previous workflow state of ``obj`` or ``None`` if no
    previous state exists.
    """
    ## We suppose that there is only one workflow.
    history = obj.workflow_history.values()[0]
    if len(history) <= 1:
        return None
    return history[-2]['review_state']


def getPreviousVersion(obj):
    """Return previous version of the object, or ``None`` if none
    could be found or if the current user is not authorized to
    retrieve it.
    """
    rtool = getToolByName(obj, 'portal_repository')
    try:
        history = rtool.getHistory(obj)
    except Unauthorized:
        return None
    if not history:
        return None
    return history[0].object



"""Define CMFNotification patches for CMFCore to call appropriated
handlers when specific events occur.

$Id$
"""

from Products.CMFCore.utils import getToolByName
from NotificationTool import ID, LOG

## FIXME: PAS does fire events, we should use them to replace these
## monkey-patches: see "PluggableAuthService/interfaces/events.py".

######### CMFCore.RegistrationTool patch ########################
def afterAdd(self, member, id, password, properties):
    """Call the original method and also CMFNotification handler."""
    self._cmf_notification_orig_afterAdd(member, id, password, properties)
    ntool = getToolByName(self, ID, None)
    if ntool is not None:
        ntool.onMemberRegistration(member, properties)

from Products.CMFCore.RegistrationTool import RegistrationTool
RegistrationTool._cmf_notification_orig_afterAdd = RegistrationTool.afterAdd
RegistrationTool.afterAdd = afterAdd
LOG.info('Monkey-patched CMFCore.RegistrationTool')
######### End of CMFCore.RegistrationTool patch #################


######### CMFCore.MemberDataTool patch ##########################
def notifyMemberModified(self):
    """Call the original method and also CMFNotification handler."""
    self._cmf_notification_orig_notifyModified()
    ntool = getToolByName(self, ID, None)
    if ntool is not None:
        membership = getToolByName(self, 'portal_membership')
        member = membership.getMemberById(self.getId())
        ntool.onMemberModification(member)

from Products.CMFCore.MemberDataTool import MemberData
MemberData._cmf_notification_orig_notifyModified = MemberData.notifyModified
MemberData.notifyModified = notifyMemberModified
LOG.info('Monkey-patched CMFCore.MemberDataTool')
######### End of CMFCore.MemberDataTool patch ###################



"""Tests for CMFNotification subscription.

$Id$
"""

from AccessControl import Unauthorized

from Products.CMFCore.utils import getToolByName

from Products.CMFNotification.exceptions import DisabledFeature
from Products.CMFNotification.NotificationTool import ID as NTOOL_ID
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION

from Products.CMFNotification.tests.base import CMFNotificationTestCase


class TestSubscription(CMFNotificationTestCase):
    """Make sure that the subscription works."""

    def afterSetUp(self):
        """Called before each tests.

        This method:

        - create users and content;

        - enable extra subscriptions.
        """
        self.createTestUsersAndContent()

        ## Enable extra subscriptions, since this is what we want to
        ## test...
        self.login('manager')
        ntool = getToolByName(self.portal, NTOOL_ID)
        ntool.manage_changeProperties(extra_subscriptions_enabled=True)

        ## Finally, log out.
        self.logout()


    def testDisabledSubscriptions(self):
        """Test if subscription is possible if it is disabled."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        document = portal.folder.document1

        self.login('manager')
        ntool.manage_changeProperties(extra_subscriptions_enabled=False)
        self.assertRaises(DisabledFeature, ntool.subscribeTo, document)

        ntool.manage_changeProperties(extra_subscriptions_enabled=True)
        ntool.subscribeTo(document) ## should not raise DisabledFeature


    def testUnauthorizedSubscriptions(self):
        """Test if subscription is possible if the user has not the
        right permission.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        document = portal.folder.document1

        self.login('manager')
        portal.manage_permission(SUBSCRIBE_PERMISSION, ('Manager', ))

        self.login('member1')
        self.assertRaises(Unauthorized, ntool.subscribeTo, document)

        portal.manage_permission(SUBSCRIBE_PERMISSION,
                                 ('Manager', 'Member'))
        ntool.subscribeTo(document) ## should not raise DisabledFeature


    def testAuthenticatedUsersSubscription(self):
        """Test subscription of authenticated users."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        document = portal.folder.document1

        ## Nodody is subscribed
        path = ntool._getPath(document)
        self.failUnlessEqual(ntool.getExtraSubscribersOf(path).keys(), [])

        ## Subscribe 'member1'
        self.login('member1')
        ntool.subscribeTo(document)
        self.failUnless(ntool.isSubscribedTo(document))

        ## Subscribe 'manager'
        self.login('manager')
        ntool.subscribeTo(document)
        self.failUnless(ntool.isSubscribedTo(document))

        ## Unsubscribe 'manager'
        ntool.unSubscribeFrom(document)
        self.failUnless(not ntool.isSubscribedTo(document))

        ## 'member1' is still subscribed
        self.login('member1')
        self.failUnless(ntool.isSubscribedTo(document))

        ## Unsubscribe 'member1'
        ntool.unSubscribeFrom(document)
        self.failUnless(not ntool.isSubscribedTo(document))

        ## Nobody is subscribed
        self.failUnlessEqual(ntool.getExtraSubscribersOf(path).keys(), [])


    def testUnSubscriptionFromObjectAbove(self):
        """Test unsubscription from the object above."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        folder = portal.folder
        document = folder.document1

        ## Subscribe to 'folder' and 'document'
        self.login('manager')
        ntool.subscribeTo(folder)
        ntool.subscribeTo(document)

        ## Unsubscribe from object above 'document' (unsubscribe from
        ## 'document' itself, actually)
        ntool.unSubscribeFromObjectAbove(document)
        self.failUnless(not ntool.isSubscribedTo(document,
                                                 as_if_not_recursive=True))
        self.failUnless(ntool.isSubscribedTo(document))
        self.failUnless(ntool.isSubscribedTo(folder))

        ## Unsubscribe from object above 'document' (unsubscribe from
        ## 'folder', actually
        ntool.unSubscribeFromObjectAbove(document)
        self.failUnless(not ntool.isSubscribedTo(document))
        self.failUnless(not ntool.isSubscribedTo(folder))

        ## Test special case where the user is subscribed to the
        ## portal root.
        ntool.subscribeTo(portal)
        self.failUnless(ntool.isSubscribedTo(portal))
        ntool.unSubscribeFromObjectAbove(folder)
        self.failUnless(not ntool.isSubscribedTo(folder))
        self.failUnless(not ntool.isSubscribedTo(portal))


    def testRecursiveSubscription(self):
        """Test recursive subscriptions."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        folder = portal.folder
        document = folder.document1

        ## Disable recursion
        ntool.manage_changeProperties(extra_subscriptions_recursive=False)
        self.login('member1')
        ntool.subscribeTo(folder)
        self.failUnless(not ntool.isSubscribedTo(document))

        ## Enable recursion
        ntool.manage_changeProperties(extra_subscriptions_recursive=True)
        self.failUnless(ntool.isSubscribedTo(document))

        ## Test 'as_if_not_recursive' parameter
        self.failUnless(not ntool.isSubscribedTo(document,
                                                 as_if_not_recursive=True))


    def testAnonymousUsersSubscription(self):
        """Test subscription of anonymous users."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        document = portal.folder.document1

        ## By default, anonymous cannot subscribe to an item
        self.logout()
        self.assertRaises(Unauthorized, ntool.subscribeTo, document)
        self.assertRaises(Unauthorized, ntool.unSubscribeFrom, document)
        self.assertRaises(Unauthorized,
                          ntool.unSubscribeFromObjectAbove, document)

        ## And if we enable anonymous subscription, it raises an
        ## error (because it is not implemented).
        portal.manage_permission(SUBSCRIBE_PERMISSION, ('Anonymous', ))
        self.logout()
        email = 'jdoe@exemple.com'
        self.assertRaises(NotImplementedError,
                          ntool.subscribeTo, document, email)
        self.assertRaises(NotImplementedError,
                          ntool.unSubscribeFrom, document, email)
        self.assertRaises(NotImplementedError,
                          ntool.unSubscribeFromObjectAbove, document, email)


    def testBrowserView(self):
        """Test subscription browser view, which is used through the
        portlet.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        folder = portal.folder
        document = folder.document1

        ## Subscribe to 'folder' and 'document'
        self.login('manager')
        document.restrictedTraverse('subscribe')()
        self.failUnless(ntool.isSubscribedTo(document))
        folder.restrictedTraverse('subscribe')()
        self.failUnless(ntool.isSubscribedTo(folder))

        ## Unsubscribe from 'document'
        document.restrictedTraverse('unsubscribe')()
        self.failUnless(not ntool.isSubscribedTo(document,
                                                 as_if_not_recursive=True))
        document.restrictedTraverse('unsubscribeFromAbove')()
        self.failUnless(not ntool.isSubscribedTo(folder))
        self.failUnless(not ntool.isSubscribedTo(document))


    def testSubscriptionToUnexpectedObject(self):
        """Test that an user is only allowed to subscribe to the
        portal object and Archetypes-based objects.
        """
        ntool = getToolByName(self.portal, NTOOL_ID)
        self.login('manager')
        isAllowed = ntool.currentUserHasSubscribePermissionOn
        self.failUnless(isAllowed(self.portal))
        self.failUnless(isAllowed(self.portal.folder))
        self.failUnless(isAllowed(self.portal.folder.document1))
        self.failUnless(not isAllowed(ntool))


class TestSubscriptionToParentAllowed(CMFNotificationTestCase):

    def afterSetUp(self):
        ntool = getToolByName(self.portal, NTOOL_ID)
        ntool.manage_changeProperties(extra_subscriptions_enabled=True,
                                      extra_subscriptions_recursive=True)
        mtool = getToolByName(self.portal, 'portal_membership')
        mtool.addMember('manager', 'manager', ['Manager'], [],
                        properties={'email': 'manager@exemple.com'})

    def _makeOne(self, context):
        from zope.component import getMultiAdapter
        from zope.component import getUtility
        from plone.portlets.interfaces import IPortletManager
        from plone.portlets.interfaces import IPortletRenderer
        from Products.CMFNotification.browser import portlet
        request = context.REQUEST
        view = context.restrictedTraverse('@@plone')
        manager = getUtility(IPortletManager,
                             name='plone.rightcolumn',
                             context=self.portal)
        assignment = portlet.Assignment()
        return getMultiAdapter(
            (context, request, view, manager, assignment),
            IPortletRenderer)

    def _callFUT(self, context):
        renderer = self._makeOne(context)
        return renderer.isSubscriptionToParentAllowed

    def test_makeOne(self):
        from Products.CMFNotification.browser.portlet import Renderer
        renderer = self._makeOne(self.portal)
        self.assert_(isinstance(renderer, Renderer))

    def test_folder_not_default_page(self):
        self.login('manager')
        self.portal.invokeFactory('Folder', 'folder')
        folder = self.portal.folder
        allowed = self._callFUT(folder)
        self.assert_(not allowed)

    def test_folder_default_page(self):
        self.login('manager')
        self.portal.invokeFactory('Folder', 'folder')
        folder = self.portal.folder
        self.portal.setDefaultPage('folder')
        allowed = self._callFUT(folder)
        self.assert_(allowed)

    def test_document_not_a_default_page(self):
        self.login('manager')
        self.portal.invokeFactory('Document', 'document')
        document = self.portal.document
        allowed = self._callFUT(document)
        self.assert_(not allowed)

    def test_document_default_page(self):
        self.login('manager')
        self.portal.invokeFactory('Document', 'document')
        document = self.portal.document
        self.portal.setDefaultPage('document')
        allowed = self._callFUT(document)
        self.assert_(allowed)

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestSubscription))
    suite.addTest(makeSuite(TestSubscriptionToParentAllowed))
    return suite



"""Define a base class and prepare ZopeTestCase and PloneTestCase to
be used in CMFNotification tests.

$Id$
"""

## Installation voodoo
from Products.Five import zcml
from Products.Five import fiveconfigure
from Products.PloneTestCase import PloneTestCase as ptc
from Products.PloneTestCase.layer import onsetup
from Testing import ZopeTestCase as ztc

@onsetup
def setup_product():
    """Yeah, you probably have seen this code a few times already..."""
    fiveconfigure.debug_mode = True
    import Products.CMFNotification
    zcml.load_config('configure.zcml', Products.CMFNotification)
    fiveconfigure.debug_mode = False
    ztc.installPackage('Five')
    ztc.installPackage('Products.CMFNotification')

setup_product()
ptc.setupPloneSite(products=['Products.CMFNotification'])


## Import statements for our own code below
from OFS.SimpleItem import SimpleItem
from Products.CMFCore.utils import getToolByName

## Monkey-patch logger so that we can test log messages (and not
## display them when tests are run).
from Products.CMFNotification.tests import fakelogger
fakelogger.patchLogging()
import logging
from Products.CMFNotification import NotificationTool
NotificationTool.LOG = logging.getLogger()


## A fake mail host which does not send emails but keep them in a
## list.
NotificationTool.MAIL_HOST_META_TYPES += ('FakeMailHost',)

class FakeMailHost(SimpleItem):
    meta_type = 'FakeMailHost'

    def __init__(self, *args, **kwargs):
        SimpleItem.__init__(self, *args, **kwargs)
        self.sent = []

    def send(self, message):
        self.sent.append(message)

    def getSentList(self):
        return self.sent

    def clearSentList(self):
        self.sent = []


class CMFNotificationBaseTestCase:
    """Base mixin for CMFNotification tests."""

    def createTestUsersAndContent(self):
        """Create some test users and content

        Three users are created:
        - a manager: 'manager'
        - two members: 'member1' and 'member2'

        The following items are created:
        + folder
        |--+-- document1
        |--+-- subfolder
        |-----+-- document2
        """
        ## Create users
        mtool = getToolByName(self.portal, 'portal_membership')
        users = ({'id': 'member1', 'roles': ('Member', )},
                 {'id': 'member2', 'roles': ('Member', )},
                 {'id': 'manager', 'roles': ('Manager', 'Member')})
        for user in users:
            email = user['id'] + '@example.com'
            mtool.addMember(user['id'], user['id'], user['roles'], [],
                            properties={'email': email})

        ## Create content
        self.login('manager')
        self.portal.invokeFactory('Folder', 'folder')
        folder = self.portal.folder
        folder.unmarkCreationFlag()
        folder.invokeFactory('Document', 'document1')
        folder.document1.unmarkCreationFlag()
        folder.invokeFactory('Folder', 'subfolder')
        subfolder = folder.subfolder
        subfolder.unmarkCreationFlag()
        subfolder.invokeFactory('Document', 'document2')
        subfolder.document2.unmarkCreationFlag()


class CMFNotificationTestCase(CMFNotificationBaseTestCase,
                              ptc.PloneTestCase):
    pass ## Nothing special to do


class CMFNotificationBrowserTestCase(CMFNotificationBaseTestCase,
                                     ptc.FunctionalTestCase):
    """A special base class for browser tests."""
    pass ## Nothing special to do.



"""A browser-view for subscriptions features.

$Id$
"""

from Products.Five import BrowserView

from Products.CMFCore.utils import getToolByName

from Products.CMFPlone import MessageFactory

from kss.core import KSSView
from kss.core import kssaction

from Products.CMFNotification.NotificationTool import ID as TOOL_ID


MF = mf = MessageFactory('cmfnotification')

class Subscription(BrowserView):
    """A browser view for subscription features.

    It is actually only a wrapper around the main methods of the tool.
    """

    def subscribe(self, subscribe_to_parent=False, notifiers=['mail']):
        item = self.context.aq_inner
        if subscribe_to_parent:
            item = item.aq_parent
            msg = MF(u'success_subscribed_parent',
                     u'You have been subscribed to the parent folder of this item.')
        else:
            msg = MF(u'success_subscribed',
                     u'You have been subscribed to this item.')

        ntool = getToolByName(item, TOOL_ID)
        ntool.subscribeTo(item, how=notifiers)

        utool = getToolByName(item, 'plone_utils')
        utool.addPortalMessage(msg)
        self.request.RESPONSE.redirect(self.context.absolute_url() + '/view')


    def unsubscribe(self):
        item = self.context.aq_inner
        ntool = getToolByName(item, TOOL_ID)
        ntool.unSubscribeFrom(item)

        msg = MF(u'success_unsubscribed',
                 u'You have been unsubscribed from this item.')
        utool = getToolByName(item, 'plone_utils')
        utool.addPortalMessage(msg)
        self.request.RESPONSE.redirect(item.absolute_url() + '/view')


    def unsubscribeFromAbove(self):
        item = self.context.aq_inner
        ntool = getToolByName(item, TOOL_ID)
        ntool.unSubscribeFromObjectAbove(item)
        msg = mf(u'success_unsubscribed_above',
                 u'You have been unsubscribed from the first parent '\
                 u'folder above.')
        utool = getToolByName(item, 'plone_utils')
        utool.addPortalMessage(msg)
        self.request.RESPONSE.redirect(item.absolute_url() + '/view')


class KSSActions(KSSView):
    """A class that holds all KSS actions related to the subscription
    portlet.
    """
    @kssaction
    def subscribe(self, portlet_hash, subscribe_to_parent=False, notifiers=['mail']):
        ## We provide a default value for 'subscribe_to_parent'
        ## because KSS will not pass it to the method is the checkbox
        ## is not checked.
        item = self.context.aq_inner
        if subscribe_to_parent:
            item = item.aq_parent
        ntool = getToolByName(item, TOOL_ID)
        ntool.subscribeTo(item, how=notifiers)
        self._refreshPortlet(portlet_hash)


    @kssaction
    def unsubscribe(self, portlet_hash):
        item = self.context.aq_inner
        ntool = getToolByName(item, TOOL_ID)
        ntool.unSubscribeFrom(item)
        self._refreshPortlet(portlet_hash)


    @kssaction
    def unsubscribeFromAbove(self, portlet_hash):
        item = self.context.aq_inner
        ntool = getToolByName(item, TOOL_ID)
        ntool.unSubscribeFromObjectAbove(item)
        self._refreshPortlet(portlet_hash)


    def _refreshPortlet(self, portlet_hash):
        commands = self.getCommandSet('plone')
        commands.refreshPortlet(portlet_hash)



"""Define CMFNotification specific permissions.

$Id$
"""
from AccessControl.SecurityInfo import ModuleSecurityInfo
from Products.CMFCore.permissions import setDefaultRoles
from config import PROJECT_NAME

security = ModuleSecurityInfo(PROJECT_NAME)

security.declarePublic('CMFNotification: Subscribe/unsubscribe')
## Warning: this value is also defined in ``profiles/default/rolemap.xml``
SUBSCRIBE_PERMISSION = 'CMFNotification: Subscribe/unsubscribe'
setDefaultRoles(SUBSCRIBE_PERMISSION, ())



"""Define CMFNotification specific exceptions.

$Id$
"""


class MailHostNotFound(Exception):
    """Could not send notification: no mailhost found"""


class DisabledFeature(Exception):
    """Cannot use this feature: it is disabled"""


class InvalidEmailAddress(Exception):
    """The given email address is not valid"""



"""Define a sub-class of ``Five.testbrowser.Browser`` that contains
useful wrappers to write functional test cases for Plone products.

FIXME: OMG, it's horrible, we have got no tests! This code is
useless... blah blah blah...

$Id$
"""

from Products.Five.testbrowser import Browser as BaseBrowser


class Browser(BaseBrowser):
    """A test browser that provides useful wrappers for Plone."""

    def submitForm(self, form_name=None, form_id=None,
                   submit_button_name=None, **kwargs):
        """Submit ``form`` with values given in ``kwargs``.

        FIXME: describe
        """
        if form_id is not None:
            form = self.getForm(id=form_id)
        elif form_name:
            form = self.getForm(name=form_name)
        else:
            raise ValueError('FIXME')
        for name, value in kwargs.items():
            control = form.getControl(name=name)
            control.value = value

        form.submit(name=submit_button_name)


    def loginAsManager(self, url=None, userid='manager', password='manager'):
        """Login to the site with the given credentials."""
        if url is not None:
            self.open(url)
        self.submitForm(form_id='login_form',
                        submit_button_name='submit',
                        __ac_name=userid,
                        __ac_password=password)


    def createItem(self, container, portal_type, **kwargs):
        """Create a new item in ``container``.

        ``kwargs`` can contain field values of the created item. For
        example: ``title``,  ``description``, etc.

        If an argument has no corresponding field, an error is raised.
        """
        url = container.absolute_url()
        url += '/createObject?type_name=%s' % portal_type
        self.open(url)
        self.submitForm(form_name='edit_form',
                        submit_button_name='form.button.save', **kwargs)


    def editItem(self, item, **kwargs):
        """Edit ``item``.

        ``kwargs`` can contain field values of the created item. For
        example: ``title``,  ``description``, etc.

        If an argument has no corresponding field, an error is raised.
        """
        url = item.absolute_url()
        url += '/edit'
        self.open(url)
        self.submitForm(form_name='edit_form',
                        submit_button_name='form.button.save', **kwargs)


    def doWorkflowTransitionOn(self, transition, item):
        """Do ``transition`` on ``item``."""
        self.open(item.absolute_url())
        ## FIXME: this a cheat
        url = item.absolute_url()
        url += '/content_status_modify?workflow_action=%s' % transition
        self.open(url)



from zope.interface import implements

from Products.Archetypes.utils import addStatusMessage
from Products.Five import BrowserView
from Products.CMFCore.utils import getToolByName
from Products.CMFNotification.browser.interfaces import \
        IUnsubscribeMenuConfiglet

class UnsubscribeMenuConfiglet(BrowserView):
    """Configlet to allow admins to unsubscribe users
    """

    implements(IUnsubscribeMenuConfiglet)

    def __init__(self, context, request):
        """
        """
        self.context = context
        self.request = request

    def getUsersList(self):
        """Get a list of users and the path for each of their subscriptions
        """
        pn = getToolByName(self, "portal_notification")
        users_list = {}
        subscriptions = pn._subscriptions
        for path in subscriptions.keys():
            user_ids = subscriptions[path]
            for user_id in user_ids:
                if user_id not in users_list.keys():
                    users_list[user_id] = [path]
                else:
                    users_list[user_id].append(path)
        return users_list

    def deleteUserOnPath(self, subscriptions, userid, path):
        subscribers = subscriptions[path]
        if userid in subscribers.keys():
            del subscribers[userid]
            subscriptions[path] = subscribers

    def unsubscribe(self):
        """init
        """
        pn = getToolByName(self, "portal_notification")
        subscriptions = pn._subscriptions
        sub_keys = subscriptions.keys()
        variables = self.request.form
        for userid in variables.keys():
            paths = variables[userid]
            for path in paths:
                if path == "all":
                    for existing_path in sub_keys:
                        self.deleteUserOnPath(subscriptions,
                                              userid,
                                              existing_path)
                else:
                    if path in sub_keys:
                        self.deleteUserOnPath(subscriptions,
                                              userid,
                                              path)
        addStatusMessage(self.request,
                         "User(s) unsubscribed successfully",
                         type='info')
        url = "%s/notification_controlpanel" % self.context.absolute_url()
        self.request.RESPONSE.redirect(url)



"""Browser tests for CMFNotification event handlers.

$Id$
"""

from Products.CMFCore.utils import getToolByName

from Products.CMFNotification.NotificationTool import ID as NTOOL_ID

from Products.CMFNotification.tests.base import FakeMailHost
from Products.CMFNotification.tests.base import CMFNotificationBrowserTestCase
from Products.CMFNotification.tests.plonetestbrowser import Browser


class TestNotificationViaBrowser(CMFNotificationBrowserTestCase):
    """Make sure that the notification works when we use a browser.

    While ``testNotification.TestNotification`` tests programmatic
    creation, edition, etc. of contents, this test case uses
    ``zope.testbrowser`` to mimic what an user does with his/her
    browser.

    Note that these tests do not check the contents of the e-mails. It
    only check the number of e-mails that have been sent.
    """

    def failUnlessSent(self, n):
        """Make the test fail unless ``n`` mails have been sent."""
        mh = self.portal.MailHost
        self.failUnless(len(mh.getSentList()) == n,
                        "Expected %s mail(s), but "\
                        "sent %s" % (n, len(mh.getSentList())))


    def afterSetUp(self):
        """Called before each tests.

        This method:

        - replaces the mailhost by a fake one which does not send
        mails but keeps them in a list;

        - create dummy users and content.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        ntool.manage_changeProperties(item_creation_notification_enabled=False)
        self.createTestUsersAndContent()
        ntool.manage_changeProperties(item_creation_notification_enabled=True)

        ## Set up a fake mail host
        if 'MailHost' in portal.objectIds():
            portal._delObject('MailHost')
        portal._setObject('MailHost', FakeMailHost())


    def testOnItemCreation(self):
        """Test notification on item creation."""
        portal = self.portal
        mh = portal.MailHost

        ## Set correct rules so that 3 mails should be sent.
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')
        changeProperty('item_creation_notification_enabled', True)
        changeProperty('on_item_creation_users', ['* :: *'])
        changeProperty('on_item_creation_mail_template',
                       ['* :: string:creation_mail_notification'])

        ## Create a document with the browser
        browser = Browser(self.portal.absolute_url() + '/login_form')
        browser.loginAsManager()
        browser.createItem(portal, 'Document', title='My document')
        self.failUnlessSent(1)
        portal.manage_delObjects(['my-document'])
        mh.clearSentList()


    ## FIXME: test "copy and paste", "cut and paste".


    def testOnItemModification(self):
        """Test notification on item modification."""
        portal = self.portal
        mh = portal.MailHost

        ## Set correct rules so that an e-mail should be sent.
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')
        changeProperty('item_modification_notification_enabled', True)
        changeProperty('on_item_modification_users', ['* :: *'])
        changeProperty('on_item_modification_mail_template',
                       ['* :: string:modification_mail_notification'])

        ## Edit a document with the browser
        browser = Browser(self.portal.absolute_url() + '/login_form')
        browser.loginAsManager()
        document = portal.folder.document1
        document.unmarkCreationFlag()
        browser.editItem(document, title='New title')
        self.failUnlessSent(1)
        mh.clearSentList()




    def testOnWorkflowTransition(self):
        """Test notification on workflow transition."""
        portal = self.portal
        mh = portal.MailHost

        ## Set correct rules so that 3 mails should be sent.
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')
        changeProperty('wf_transition_notification_enabled', True)
        changeProperty('on_wf_transition_users',
                       ['* :: *'])
        changeProperty('on_wf_transition_mail_template',
                       ['* :: string:workflow_mail_notification'])

        ## Publish item with the browser
        browser = Browser(self.portal.absolute_url() + '/login_form')
        browser.loginAsManager()
        document = portal.folder.document1
        browser.doWorkflowTransitionOn('publish', document)
        self.failUnlessSent(3)
        mh.clearSentList()


    def testOnMemberRegistration(self):
        """Test notification on member registration."""
        portal = self.portal
        mh = portal.MailHost

        ## Set correct rules so that 3 mails should be sent.
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')
        changeProperty('member_registration_notification_enabled', True)
        changeProperty('on_member_registration_users', ['* :: *'])
        changeProperty('on_member_registration_mail_template',
                       ['* :: string:registration_mail_notification'])

        ## Register new member with the browser
        ## FIXME
        #userid = 'a_new_member'
        #self.failUnlessSent(3)
        #mtool = getToolByName(portal, 'portal_membership')
        #mtool.deleteMembers([userid, ])
        mh.clearSentList()


    def testOnMemberModification(self):
        """Test notification on member modification."""
        pass ## FIXME: implement me! (Damien)


    def testOnDiscussionItemCreation(self):
        """Test notification on discussion item creation."""
        portal = self.portal

        ## Set correct rules so that 1 e-mail should be sent.
        ntool = getToolByName(portal, NTOOL_ID)
        changeProperty = lambda key, value: \
            ntool.manage_changeProperties(**{key: value})
        self.login('manager')
        changeProperty('on_discussion_item_creation_notification_enabled',
                       True)
        changeProperty('on_discussion_item_creation_users',
                       ['* :: *'])
        changeProperty('on_discussion_item_creation_mail_template',
                       ['* :: string:discussion_mail_notification'])

        ## Add discussion item with the browser.
        document = portal.folder.document1
        document.allow_discussion = True
        ## FIXME
        #self.failUnlessSent(1)
        #mh.clearSentList()


    def testHandlersDoNotOverlap(self):
        """Test that handlers do not overlap themselves, i.e. that at
        most one handler is called for any event.
        """
        portal = self.portal
        mh = portal.MailHost

        ## Enable everything
        self.login('manager')
        users = ['* :: *']
        properties = {
           'item_creation_notification_enabled': True,
           'on_item_creation_users': users,
           'on_item_creation_mail_template': ['* :: string:creation_mail_notification'],
           'item_modification_notification_enabled': True,
           'on_item_modification_users': users,
           'on_item_modification_mail_template': ['* :: string:modification_mail_notification'],
           'wf_transition_notification_enabled': True,
           'on_wf_transition_users': users,
           'on_wf_transition_mail_template': ['* :: string:workflow_mail_notification'],
           'member_registration_notification_enabled': True,
           'on_member_registration_users': users,
           'on_member_registration_mail_template':['* :: string:registration_mail_notification'],
           'discussion_item_creation_notification_enabled': True,
           'on_discussion_item_creation_users': users,
           'on_discussion_item_creation_mail_template': ['* :: string:discussion_mail_notification'],
           }
        ntool = getToolByName(portal, NTOOL_ID)
        ntool.manage_changeProperties(properties)

        ## Initialize browser and log in
        browser = Browser(self.portal.absolute_url() + '/login_form')
        browser.loginAsManager()

        ## Create item
        browser.createItem(portal, 'Document', title='My document')
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Modify it
        document = portal['my-document']
        browser.editItem(document, title='New title')
        self.failUnlessSent(1)
        mh.clearSentList()

        ## Publish it
        browser.doWorkflowTransitionOn('publish', document)
        self.failUnlessSent(3)
        mh.clearSentList()

        ## Test member registration
        #userid = 'a_new_member'
        ## FIXME: register new member
        #self.failUnlessSent(3)
        #mh.clearSentList()

        ## FIXME: Test member modification

        ## Test discussion item creation
        #document.allow_discussion = True
        ## FIXME: add discussion item
        #self.failUnlessSent(1)
        #mh.clearSentList()


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestNotificationViaBrowser))
    return suite



"""Product initialization.

$Id$
"""
from Products.CMFCore import utils as CMFCoreUtils

## Apply monkey patches and set permissions
import Products.CMFNotification.patches
import Products.CMFNotification.permissions

def initialize(context):
    from Products.CMFNotification import NotificationTool
    tools = (NotificationTool.NotificationTool, )
    CMFCoreUtils.ToolInit(NotificationTool.META_TYPE,
                          tools=tools,
                          icon='tool.gif').initialize(context)



# This file is part of CMFNotification
#
# Copyright (c) 2005-2008 by Pilot Systems (http://www.pilotsystems.net)
# 
# CMFNotification is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
"""Define subscription portlet.

$Id$
"""

from zope.interface import implements
from zope.component import getMultiAdapter
from zope.component import getUtilitiesFor
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

from Products.CMFCore.utils import getToolByName

from Products.CMFPlone import MessageFactory
from plone.app.portlets.portlets import base
from plone.portlets.interfaces import IPortletDataProvider

from Products.CMFNotification.NotificationTool import ID as NTOOL_ID
from Products.CMFNotification.interfaces import INotificationDelivery


mf = MessageFactory('cmfnotification')


class ISubscriptionPortlet(IPortletDataProvider):
    """A portlet which allow an user to subscribe/unsubscribe to
    content changes.
    """


class Assignment(base.Assignment):
    implements(ISubscriptionPortlet)

    title = mf(u'notification_portlet_header',
               default=u'Subscription')


class Renderer(base.Renderer):

    render = ViewPageTemplateFile('templates/portlet.pt')


    def __init__(self, context, request, view, manager, data):
        base.Renderer.__init__(self, context, request, view, manager, data)
        ## FIXME: looks like 'update()' is not called before render if
        ## I do not do that
        if self.available:
            self.update()


    @property
    def available(self):
        ntool = getToolByName(self.context, NTOOL_ID, None)
        if ntool is None:
            ## Looks like the product has been removed but the portlet
            ## is still configured to appear somewhere.
            return False
        if not ntool.isExtraSubscriptionsEnabled():
            return False

        # Don't show the portlet for temporary content in portal_factory
        if getattr(self.context, 'isTemporary', lambda: False)():
            return False

        return ntool.currentUserHasSubscribePermissionOn(self.context)


    @property
    def notifiers(self):
        li = [dict(id=i, description=d.description) for \
            i, d in getUtilitiesFor(INotificationDelivery)]
        return li

    @property
    def isSubscriptionToParentAllowed(self):
        """Return whether subscription to the parent of the context
        (i.e. the first folderish item above the context) is allowed.
        """
        ntool = getToolByName(self.context, NTOOL_ID)
        if not ntool.isExtraSubscriptionsRecursive():
            return False

        context = self.context.aq_inner
        plone_view = getMultiAdapter((context, self.request),
                                     name='plone')

        if not plone_view.isDefaultPageInFolder():
            return False

        parent = context.aq_parent
        return ntool.currentUserHasSubscribePermissionOn(parent)


    def update(self):
        ntool = getToolByName(self.context, NTOOL_ID)
        self.is_subscribed = ntool.isSubscribedTo(self.context)
        self.is_subscribed_here = ntool.isSubscribedTo(self.context,
                                                       as_if_not_recursive=True)


class AddForm(base.NullAddForm):
    def create(self):
        return Assignment()



"""Tests for CMFNotification subscription.

$Id: testSubscription.py 229875 2011-01-06 10:42:55Z dbaty $
"""

from Products.CMFCore.utils import getToolByName

from Products.CMFNotification.NotificationTool import ID as NTOOL_ID

from Products.CMFNotification.tests.base import CMFNotificationTestCase


class TestUnsubscribeMenuConfiglet(CMFNotificationTestCase):
    """Make sure that the Unsubscribe Menu configlet is working
    as expected"""

    def afterSetUp(self):
        """Called before each tests.

        This method:
        - Creates users and content;
        - Enables extra subscriptions
        - Subscribes the users to some objects
        """
        self.createTestUsersAndContent()

        self.login('manager')
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)
        ntool.manage_changeProperties(extra_subscriptions_enabled=True)

        folder = portal.folder
        document = folder.document1

        ## Subscribe 'member1'
        self.login('member1')
        ntool.subscribeTo(document)

        ## Subscribe 'manager'
        self.login('manager')
        ntool.subscribeTo(document)
        ntool.subscribeTo(folder)

    def testGetUsersList(self):
        """Test that we get the proper list of users subscribed"""
        unsubscribemenu_configlet = self.portal.unrestrictedTraverse(
            "@@unsubscribemenu_configlet"
        )
        users_list = unsubscribemenu_configlet.getUsersList()
        expected_users_list = {
            'manager': ['/folder/', '/folder/document1'],
            'member1': ['/folder/document1']
        }
        for user in users_list.keys():
            paths = users_list[user]
            self.assertTrue(user in expected_users_list.keys())
            self.failUnless(paths == expected_users_list[user])

    def testUnsubscribe(self):
        """Test unsubscribing manager and member1 from the document1"""
        unsubscribemenu_configlet = self.portal.unrestrictedTraverse(
            "@@unsubscribemenu_configlet"
        )
        # make fake request
        unsubscribemenu_configlet.request.form = {
            'manager': ["/folder/document1"],
            "member1": ["/folder/document1"]
        }
        unsubscribemenu_configlet.unsubscribe()
        users_list = unsubscribemenu_configlet.getUsersList()
        expected_users_list = {
            'manager': ['/folder/'],
        }
        for user in users_list.keys():
            paths = users_list[user]
            self.assertTrue(user in expected_users_list.keys())
            self.failUnless(paths == expected_users_list[user])


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestUnsubscribeMenuConfiglet))
    return suite




"""Tests for CMFNotification installation ad uninstallation.

$Id$
"""

from zope.component import getUtility
from zope.component import getMultiAdapter
from AccessControl.PermissionRole import rolesForPermissionOn

from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletAssignmentMapping

from Products.CMFCore.utils import getToolByName

from Products.CMFNotification.config import LAYER_NAME
from Products.CMFNotification.config import PORTLET_NAME
from Products.CMFNotification.NotificationTool import ID as TOOL_ID
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION

from Products.CMFNotification.tests.base import CMFNotificationTestCase


class TestInstallation(CMFNotificationTestCase):
    """Make sure that the product is properly installed."""

    def afterSetUp(self):
        pass

    def testToolIsThere(self):
        tool = getToolByName(self.portal, TOOL_ID)
        self.failUnless(tool is not None)

    def testSkinLayerIsThere(self):
        stool = getToolByName(self.portal, 'portal_skins')
        for skin, layers in stool._getSelections().items():
            layers = layers.split(',')
            self.failUnless(LAYER_NAME in layers)
        self.failUnless(LAYER_NAME in stool.objectIds())

    def testPortletCanBeAdded(self):
        for name in ('plone.leftcolumn', 'plone.rightcolumn'):
            manager = getUtility(IPortletManager,
                                 name=name,
                                 context=self.portal)
            titles = [p.title for p in manager.getAddablePortletTypes()]
            self.failUnless(PORTLET_NAME in titles)

        manager = getUtility(IPortletManager,
                             name='plone.rightcolumn',
                             context=self.portal)
        right_portlets = getMultiAdapter((self.portal, manager),
                                         IPortletAssignmentMapping,
                                         context=self.portal)
        right_portlets = right_portlets.keys()
        self.failUnless(PORTLET_NAME in right_portlets)

    def testPermissionHasBeenSet(self):
        roles = set(rolesForPermissionOn(SUBSCRIBE_PERMISSION, self.portal))
        self.failUnlessEqual(roles, set(('Manager', 'Member')))

    def testConfigurationConfigletHasBeenAdded(self):
        cptool = getToolByName(self.portal, 'portal_controlpanel')
        configlets = [c.getId() for c in cptool.listActions()]
        self.failUnless('cmfnotification_configuration' in configlets)

    def testUnsubscribeConfigletHasBeenAdded(self):
        cptool = getToolByName(self.portal, 'portal_controlpanel')
        configlets = [c.getId() for c in cptool.listActions()]
        self.failUnless('cmf_notification_unsubscribemenu' in configlets)


class TestUnInstallation(CMFNotificationTestCase):
    """Test that the product has been properly uninstalled."""

    def afterSetUp(self):
        """Uninstall the product before running each test."""
        qtool = getToolByName(self.portal, 'portal_quickinstaller')
        self.setRoles(['Manager'])
        products = self.portal.Control_Panel.Products
        version = \
            products.get('CMFQuickInstallerTool', {}).get('version', '')
        version = version.split('.')[0]
        ## Plone 3.x is shipped with CMFQuickInstallerTool 2.1.7 which
        ## does not see Products.CMFNotification in the list of
        ## uninstallable products. Plone 4.x ships with version 3
        ## where the bug has been fixed (see changeset 110252).
        if version == '2':
            from Products.CMFNotification.Extensions.Install import uninstall
            uninstall(self.portal)
        else:
            qtool.uninstallProducts(['CMFNotification'])

    def testToolIsNotThere(self):
        tool = getToolByName(self.portal, TOOL_ID, None)
        self.failUnless(tool is None)

    def testSkinLayerIsNotThere(self):
        stool = getToolByName(self.portal, 'portal_skins')
        for skin, layers in stool._getSelections().items():
            layers = layers.split(',')
            self.failUnless(LAYER_NAME not in layers)
        self.failUnless(LAYER_NAME not in stool.objectIds())

    def testPortletDoNoExist(self):
        for name in ('plone.leftcolumn', 'plone.rightcolumn'):
            manager = getUtility(IPortletManager,
                                 name=name,
                                 context=self.portal)
            titles = [p.title for p in manager.getAddablePortletTypes()]
            self.failUnless(PORTLET_NAME not in titles)

        manager = getUtility(IPortletManager,
                             name='plone.rightcolumn',
                             context=self.portal)
        right_portlets = getMultiAdapter((self.portal, manager),
                                         IPortletAssignmentMapping,
                                         context=self.portal)
        right_portlets = right_portlets.keys()
        self.failUnless(PORTLET_NAME not in right_portlets)

    def testConfigurationConfigletDoNotExist(self):
        cptool = getToolByName(self.portal, 'portal_controlpanel')
        configlets = [c.getId() for c in cptool.listActions()]
        self.failUnless('cmfnotification_configuration' not in configlets)

    def testUnsubscribeConfigletDoNotExist(self):
        cptool = getToolByName(self.portal, 'portal_controlpanel')
        configlets = [c.getId() for c in cptool.listActions()]
        self.failUnless('cmf_notification_unsubscribemenu' not in configlets)

def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestInstallation))
    suite.addTest(makeSuite(TestUnInstallation))
    return suite



"""Miscellaneous tests for CMFNotification.

$Id$
"""

from zope.component import getUtility
from zope.component import getMultiAdapter

from Products.CMFCore.utils import getToolByName

from plone.portlets.interfaces import IPortletManager
from plone.portlets.interfaces import IPortletRenderer

from Products.CMFNotification.utils import getPreviousVersion
from Products.CMFNotification.utils import getPreviousWorkflowState
from Products.CMFNotification.NotificationTool import ID as NTOOL_ID
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION
from Products.CMFNotification.tests.base import CMFNotificationTestCase
from Products.CMFNotification.browser import portlet as CMFNotificationPortlet


class TestMisc(CMFNotificationTestCase):
    """Test various utility methods."""

    def afterSetUp(self):
        """Create dummy content."""
        self.createTestUsersAndContent()


    def testGetPath(self):
        """Test ``NotificationTool._getPath()``."""
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)

        getPath = ntool._getPath
        for obj, path in (
            (portal, '/'),
            (portal.folder, '/folder/'),
            (portal.folder.document1, '/folder/document1'),
            (portal.folder.subfolder, '/folder/subfolder/'),
            (portal.folder.subfolder.document2, '/folder/subfolder/document2'),
            ):
            self.failUnlessEqual(getPath(obj), path)


    def testGetParents(self):
        """Test ``NotificationTool._getParents()``."""
        ntool = getToolByName(self.portal, NTOOL_ID)

        getParents = ntool._getParents
        for path, parents in (
            ('/', []),
            ('/folder/', ('/', )),
            ('/folder/document1', ('/folder/', '/')),
            ('/folder/subfolder/', ('/folder/', '/')),
            ('/folder/subfolder/document2',
             ('/folder/subfolder/', '/folder/', '/')),
            ):
            self.failUnlessEqual(getParents(path), parents)


    def testIsSubscriptionToParentAllowed(self):
        """Test ``NotificationTool.isSubscriptionToParentAllowed()``.
        """
        portal = self.portal
        ntool = getToolByName(portal, NTOOL_ID)

        def isAllowed(obj):
            plone_view = self.folder.restrictedTraverse('@@plone')
            manager = getUtility(IPortletManager,
                                 name='plone.rightcolumn',
                                 context=obj)
            ## Remove previously cached view.
            if hasattr(obj.REQUEST, '__annotations__'):
                del obj.REQUEST.__annotations__
            renderer = getMultiAdapter((obj, obj.REQUEST,
                                        plone_view, manager,
                                        CMFNotificationPortlet.Assignment()),
                                       IPortletRenderer)
            return renderer.isSubscriptionToParentAllowed

        ntool.manage_changeProperties(extra_subscriptions_enabled=True)
        ntool.manage_changeProperties(extra_subscriptions_recursive=True)
        self.failUnless(not isAllowed(portal.folder))
        self.failUnless(not isAllowed(portal.folder.document1))

        portal.folder.setDefaultPage('document1')

        portal.manage_permission(SUBSCRIBE_PERMISSION, ())
        self.failUnless(not isAllowed(portal.folder.document1))

        portal.manage_permission(SUBSCRIBE_PERMISSION, ('Manager', ))
        self.failUnless(isAllowed(portal.folder.document1))

        ntool.manage_changeProperties(extra_subscriptions_recursive=False)
        self.failUnless(not isAllowed(portal.folder.document1))


    def testGetPreviousVersion(self):
        """Test ``utils._getPreviousVersion()``."""
        document = self.portal.folder.document1
        ## At this point, the versioning mechanism does not yet hold
        ## any reference to 'document'. Even this first version is not
        ## registered, because we have only called 'invokeFactory()'
        ## to create dummy items...
        previous = getPreviousVersion(document)
        self.failUnlessEqual(previous, None)

        ## ... We have to register it manually to mimic what normally
        ## happens through the web.
        rtool = getToolByName(self.portal, 'portal_repository')
        rtool.save(document, comment='initial version')

        title = document.Title()
        document.setTitle('New title')
        document.update_version_on_edit() ## Mimic TTW edition
        previous = getPreviousVersion(document)
        self.failUnlessEqual(document.Title(), 'New title')
        self.failUnlessEqual(previous.Title(), title)


    def testGetPreviousWorkflowState(self):
        """Test ``utils.getPreviousWorkflowState()``."""
        wtool = getToolByName(self.portal, 'portal_workflow')
        document = self.portal.folder.document1
        self.failUnlessEqual(getPreviousWorkflowState(document), None)
        state = wtool.getInfoFor(document, 'review_state')
        wtool.doActionFor(document, 'publish')
        self.failUnlessEqual(getPreviousWorkflowState(document), state)


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestMisc))
    return suite



## Copyright (c) 2007-2008 Damien Baty
##
## This file is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 3 of the License,
## or (at your option) any later version.
##
## This file is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

"""Define ``FakeLogger`` and ``FakeRootLogger`` classes, which can be
used to monkey-patch the corresponding classes of the ``logging``
package so that every messages is stored in the logger itself and can
then be checked against what was expected to be logged.

Basic usage
===========

Suppose that we want to test this method::

    >>> import logging
    >>> def foo():
    ...     ## Do things
    ...     logging.warning("OMG, we've done horrible things!")

We can probably test that things have been done, but we may also want
to check that the warning has been logged, too. To do that, we first
call the ``patch`` method::

    >>> from fakelogger import patchLogging
    >>> patchLogging()

Now we have a fake logger, which do not actually write anything to a
file or to the standard error. It has additional methods with which we
can test that our method has actually logged what we expected it to::

    >>> foo()
    >>> from logging import getLogger
    >>> logger = getLogger()
    >>> logger.getStack()
    ["OMG, we've done horrible things!"]

Obviously, as the name of the method suggests, the logger has a
stack. So if we call our method again, we will have the warning
twice::

    >>> foo()
    >>> logger.getStack()
    ["OMG, we've done horrible things!", "OMG, we've done horrible things!"]

That can be quite handy if we have to log multiple things in an
integration test, for example. However, for unit tests, it is often
useful to clear the stack::

    >>> logger.clearStack()
    >>> logger.getStack()
    []

That is why the "natural" usage could be::

    >>> logger.clearStack() ## You never know...
    >>> foo()
    >>> logger.getStack() ## The test itself
    ["OMG, we've done horrible things!"]
    >>> logger.clearStack() ## Be kind to other test(er)s

Note that in the previous example, we have used the root logger by
using ``getLogger`` without any parameter. We can use any logger,
though::

    >>> logger = getLogger('My own logger')
    >>> def bar():
    ...     ## Do other horrible things
    ...     logger.error('We have killed Kenny.')
    >>> bar()
    >>> logger.getStack()
    ['We have killed Kenny.']


Advanced usage
==============

There is no advanced usage, yet. See FIXME below.


$Id$
"""

## FIXME: could we not use 'MemoryHandler'?

import logging
from logging import Logger
from logging import RootLogger
from logging import WARNING


def patchLogging():
    logging._loggerClass = FakeLogger
    logging.root = FakeRootLogger(WARNING)


class FakeLogger(Logger):
    def _log(self, level, msg, args, exc_info=None):
        """Store ``msg`` in a stack."""
        stack = self.getStack()
        stack.append(msg % args)
        ## FIXME: this is a bit basic, for now:
        ## - what if exc_info is True?
        ## - do we want to check log records level (severity)?


    def getStack(self):
        if not hasattr(self, '_stack'):
            self.clearStack()
        return self._stack


    def clearStack(self):
        self._stack = []


class FakeRootLogger(FakeLogger, RootLogger):
    pass



def _test():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _test()



import logging
from ZODB.POSException import ConflictError
from zope.interface import implements
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import MessageFactory
from Products.CMFNotification.utils import encodeMailHeaders
from Products.CMFNotification import NotificationTool
from Products.CMFNotification.exceptions import MailHostNotFound
from Products.CMFNotification.interfaces import INotificationDelivery

LOG = logging.getLogger('CMFNotification.mail')

_ = MessageFactory('cmfnotification')

class MailNotificationDelivery(object):
    implements(INotificationDelivery)

    @property
    def description(self):
        return _(u'mail_notification_delivery_description',
                   default=u'Notify by email')

    def notify(self, obj, user, what, label, bindings):
        mtool = getToolByName(obj, 'portal_membership')
        member = mtool.getMemberById(str(user))
        if member is None:
            return 0
        email = member.getProperty('email', '')
        if not email:
            return 0
        if not NotificationTool.EMAIL_REGEXP.match(email):
            return 0

        pn = getToolByName(obj, 'portal_notification')
        template = pn.getTemplate(obj, what, bindings)
        if template is None:
            LOG.warning("No mail template for label '%s' for "\
                        "'%s' notification of '%s'",
                        label, what, obj.absolute_url(1))
            return 0

        try:
            message = template(**bindings)
        except ConflictError:
            raise
        except:
            LOG.error("Cannot evaluate mail template '%s' on '%s' "\
                      "for '%s' for label '%s'",
                      template.absolute_url(1),
                      obj.absolute_url(1), what, label,
                      exc_info=True)
            return 0
        return self.sendNotification(obj, email, message)

    def sendNotification(self, obj, address, message):
        """Send ``message`` to ``address``."""
        portal = obj.restrictedTraverse('@@plone_portal_state').portal()
        mailhosts = portal.superValues(NotificationTool.MAIL_HOST_META_TYPES)
        if not mailhosts:
            raise MailHostNotFound
        mailhost = mailhosts[0]

        ptool = getToolByName(obj, 'portal_properties').site_properties
        encoding = ptool.getProperty('default_charset', 'utf-8')
        message = encodeMailHeaders(message, encoding)

        this_message = ('To: %s\n' % address) + message
        this_message = this_message.encode(encoding)
        try:
            mailhost.send(this_message)
        except ConflictError:
            raise
        except:
            LOG.error('Error while sending '\
                      'notification: \n%s' % this_message,
                      exc_info=True)
            return 0
        return 1



from zope.interface import Interface

class IUnsubscribeMenuConfiglet(Interface):
    """Configlet to allow admins to unsubscribe users
    """

    def getUsersList():
        """Get a list of users and the path for each of their subscriptions
        """

    def unsubscribe():
        """init
        """




import os
from setuptools import find_packages
from setuptools import setup

version = '2.3b3'
description = open('README.txt').read().strip()
long_description = '\n\n'.join((
        open(os.path.join(
        'Products', 'CMFNotification', 'README.txt')).read().strip(),
        open(os.path.join("docs", "HISTORY.txt")).read().strip()))

setup(
    name='Products.CMFNotification',
    version=version,
    description=description,
    long_description=long_description,
    classifiers=(
        'Framework :: Plone',
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        ),
    keywords='CMFNotification plone notification e-mail',
    author='Pilot Systems',
    author_email='ploneorg@pilotsystems.net',
    url='http://plone.org/products/cmfnotification',
    license='GPL',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['Products'],
    include_package_data=True,
    zip_safe=False,
    install_requires=(
        'setuptools',
        'Plone >= 3.0',
        'plone.app.kss',
        ),
    extras_require={
        'test': [
            'Products.PloneTestCase',
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
    )



# This file is part of CMFNotification
#
# Copyright (c) 2005-2008 by Pilot Systems (http://www.pilotsystems.net)
#
# CMFNotification is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
"""Define notification tool.

$Id$
"""

import re
import inspect
import logging
from types import StringType

from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from ZODB.POSException import ConflictError
from OFS.PropertyManager import PropertyManager
from BTrees.IOBTree import IOBTree
from BTrees.OOBTree import OOBTree
from zope.component import getUtility

from AccessControl import Unauthorized
from AccessControl import ClassSecurityInfo
from AccessControl import getSecurityManager
from AccessControl.PermissionRole import rolesForPermissionOn

from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.Expression import Expression
from Products.CMFEditions.interfaces.IArchivist import ArchivistUnregisteredError

from Products.CMFNotification.utils import getBasicBindings
from Products.CMFNotification.utils import getPreviousVersion
from Products.CMFNotification.utils import getExpressionContext
from Products.CMFNotification.utils import getPreviousWorkflowState
from Products.CMFNotification.exceptions import DisabledFeature
from Products.CMFNotification.exceptions import InvalidEmailAddress
from Products.CMFNotification.permissions import SUBSCRIBE_PERMISSION
from Products.CMFNotification.interfaces import INotificationDelivery

from Products.ATContentTypes.interface.interfaces import IATContentType
from Products.CMFPlone.interfaces import IPloneSiteRoot

from plone.dexterity.interfaces import IDexterityContent


ID = 'portal_notification'
TITLE = 'CMF Notification tool'
META_TYPE = 'CMFNotificationTool'

## This regexp was taken from Silva ('Silva/subscriptionservice.py').
EMAIL_REGEXP = re.compile('^[0-9a-zA-Z_&.%+-]+@([0-9a-zA-Z]([0-9a-zA-Z-]*[0-9a-zA-Z])?\.)+[a-zA-Z]{2,6}$')
RULE_DELIMITER = '::'
LOG = logging.getLogger('CMFNotification')
## Ignore objects temporarily created by the Plone Factory Tool
DEFAULT_IGNORE_RULES = ["python: getattr(context, 'isTemporary', lambda: False)()"]

MAIL_HOST_META_TYPES = ('Mail Host', 'Secure Mail Host', 'Maildrop Host',
    'Secure Maildrop Host')


class NotificationTool(UniqueObject, SimpleItem, PropertyManager):
    """Main notification tool."""
    id = ID
    title = TITLE
    meta_type = META_TYPE

    manage_options = (PropertyManager.manage_options
                      + SimpleItem.manage_options)

    ## Extra subscriptions
    extra_subscriptions_enabled = False
    extra_subscriptions_recursive = True

    ## Debug mode
    debug_mode = False

    ## Ignore rules
    ignore_rules = DEFAULT_IGNORE_RULES

    ## Item creation
    item_creation_notification_enabled = False
    on_item_creation_users = []
    on_item_creation_mail_template = []
    ## Item modification
    item_modification_notification_enabled = False
    on_item_modification_users = []
    on_item_modification_mail_template = []
    ## Item removal
    item_removal_notification_enabled = False
    on_item_removal_users = []
    on_item_removal_mail_template = []
    ## Workflow transition
    wf_transition_notification_enabled = False
    on_wf_transition_users = []
    on_wf_transition_mail_template = []
    ## Member registration
    member_registration_notification_enabled = False
    on_member_registration_users = []
    on_member_registration_mail_template = []
    ## Member modification
    member_modification_notification_enabled = False
    on_member_modification_users = []
    on_member_modification_mail_template = []
    ## Discussion item creation
    discussion_item_creation_notification_enabled = False
    on_discussion_item_creation_users = []
    on_discussion_item_creation_mail_template = []

    _properties = ({'id': 'extra_subscriptions_enabled',
                    'label': 'Enable extra subscriptions',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'extra_subscriptions_recursive',
                    'label': 'Toggle recursive mode for extra subscriptions',
                    'mode': 'w',
                    'type': 'boolean'},

                   {'id': 'debug_mode',
                    'label': 'Toggle debug mode',
                    'mode': 'w',
                    'type': 'boolean'},

                   {'id': 'ignore_rules',
                    'label': 'Rules (ignore)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'item_creation_notification_enabled',
                    'label': 'Enable item creation notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_item_creation_users',
                    'label': 'Rules on item creation (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_item_creation_mail_template',
                    'label': 'Rules on item creation (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'item_modification_notification_enabled',
                    'label': 'Enable item modification notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_item_modification_users',
                    'label': 'Rules on item modification (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_item_modification_mail_template',
                    'label': 'Rules on item modification (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'item_removal_notification_enabled',
                    'label': 'Enable item removal notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_item_removal_users',
                    'label': 'Rules on item removal (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_item_removal_mail_template',
                    'label': 'Rules on item removal (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'wf_transition_notification_enabled',
                    'label': 'Enable workflow transition notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_wf_transition_users',
                    'label': 'Rules on workflow transition (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_wf_transition_mail_template',
                    'label': 'Rules on workflow transition (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'member_registration_notification_enabled',
                    'label': 'Enable member registration notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_member_registration_users',
                    'label': 'Rules on member registration (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_member_registration_mail_template',
                    'label': 'Rules on member registration (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'member_modification_notification_enabled',
                    'label': 'Enable member modification notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_member_modification_users',
                    'label': 'Rules on member modification (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_member_modification_mail_template',
                    'label': 'Rules on member modification (mail template)',
                    'mode': 'w',
                    'type': 'lines'},

                   {'id': 'discussion_item_creation_notification_enabled',
                    'label': 'Enable discussion item creation notification',
                    'mode': 'w',
                    'type': 'boolean'},
                   {'id': 'on_discussion_item_creation_users',
                    'label': 'Rules on discussion item creation (users)',
                    'mode': 'w',
                    'type': 'lines'},
                   {'id': 'on_discussion_item_creation_mail_template',
                    'label': 'Rules on discussion item creation (mail template)',
                    'mode': 'w',
                    'type': 'lines'},
                   )

    security = ClassSecurityInfo()
    decPrivate = security.declarePrivate
    decProtected = security.declareProtected
    decPublic = security.declarePublic


    def __init__(self, *args, **kwargs):
        self._uid_to_path = IOBTree()
        self._subscriptions = OOBTree()


    #################################################################
    ## Notification handlers
    ########################
    decPrivate('onItemCreation')
    def onItemCreation(self, obj):
        """Handler called when an item is created.

        It returns the number of mails which have been sent.

        **Warning:** this handler is not called when a discussion item
        is added. In this case, ``onDiscussionItemCreation()`` is
        called instead.
        """
        if not self.getProperty('item_creation_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        extra_bindings = getBasicBindings(obj)
        changenote = self._getChangeNote(obj)
        extra_bindings.update({'changenote': changenote })
        return self._handlerHelper(obj, 'item_creation', extra_bindings)


    decPrivate('onItemModification')
    def onItemModification(self, obj):
        """Handler called when an item is modified.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('item_modification_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        extra_bindings = getBasicBindings(obj)
        changenote = self._getChangeNote(obj)
        extra_bindings.update({'current': obj,
                               'previous': getPreviousVersion(obj),
                               'changenote': changenote })
        return self._handlerHelper(obj, 'item_modification', extra_bindings)

    decPrivate('onItemRemoval')
    def onItemRemoval(self, obj):
        """Handler called when an item is removed.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('item_removal_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        extra_bindings = getBasicBindings(obj)
        return self._handlerHelper(obj, 'item_removal', extra_bindings)


    decPrivate('onWorkflowTransition')
    def onWorkflowTransition(self, obj, action):
        """Handler called when a workflow transition is triggered.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('wf_transition_notification_enabled'):
            return 0
        if self.ignoreNotification(obj):
            return 0

        wtool = getToolByName(self, 'portal_workflow')
        comments = wtool.getInfoFor(obj, 'comments')
        extra_bindings = getBasicBindings(obj)
        extra_bindings.update({'transition': action,
                               'comments': comments,
                               'previous_state': getPreviousWorkflowState(obj)})
        return self._handlerHelper(obj, 'wf_transition', extra_bindings)


    decPrivate('onMemberRegistration')
    def onMemberRegistration(self, member, properties):
        """Handler called when a new portal member has been
        registered.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('member_registration_notification_enabled'):
            return 0
        if self.ignoreNotification(member):
            return 0

        if properties is None:
            properties = {} ## FIXME: How could it be? (Damien)

        current_user = getSecurityManager().getUser()
        extra_bindings = getBasicBindings(member)
        extra_bindings.update({'current_user': current_user,
                               'member': member,
                               'properties': properties,
                               'event': 'registration'})
        return self._handlerHelper(member, 'member_registration',
                                   extra_bindings)


    decPrivate('onMemberModification')
    def onMemberModification(self, member):
        """Handler called when a member changes his/her properties.

        It returns the number of mails which have been sent.
        """
        ## FIXME: this should go away when we rely on the appropriate
        ## event.
        ## This method can also be called when the member is
        ## registered. We have to check that.
        stack = inspect.stack()
        ## 1st item is ourself
        ## 2nd item is 'CMFCore.MemberDataTool.notifyMemberModified()'
        ## 3rd item is 'CMFCore.MemberDataTool.setMemberProperties()'
        ## 4th item is what we want to check: it is either 'addMember'
        ## or 'setProperties()'
        caller = stack[3][3]
        if caller != 'setProperties':
            return 0

        if not self.getProperty('member_modification_notification_enabled'):
            return 0
        if self.ignoreNotification(member):
            return 0

        ## FIXME: what is the purpose of the following lines? (Damien)
        memberdata = getToolByName(self, 'portal_memberdata')
        properties = {}
        for key, value in memberdata.propertyItems():
            properties[key] = value

        current_user = getSecurityManager().getUser()
        extra_bindings = getBasicBindings(None)
        extra_bindings.update({'current_user': current_user,
                               'member': member,
                               'properties': properties,
                               'event': 'modification'})
        return self._handlerHelper(member, 'member_modification',
                                   extra_bindings)


    decPrivate('onDiscussionItemCreation')
    def onDiscussionItemCreation(self, discussion_item):
        """Handler called when a discussion item is created.

        It returns the number of mails which have been sent.
        """
        if not self.getProperty('discussion_item_creation_notification_enabled'):
            return 0
        if self.ignoreNotification(discussion_item):
            return 0

        ## We add two bindings to disambiguate the meaning of 'here'
        ## in the mail template and the rules: 'discussion_item' and
        ## 'discussed_item'.
        discussed_item = discussion_item
        while discussed_item.meta_type == discussion_item.meta_type:
            discussed_item = discussed_item.aq_inner.aq_parent.aq_parent
        extra_bindings = getBasicBindings(discussed_item)
        extra_bindings.update({'discussion_item': discussion_item,
                               'discussed_item': discussed_item})
        return self._handlerHelper(discussion_item, 'discussion_item_creation',
                                   extra_bindings)


    def _handlerHelper(self, obj, what, extra_bindings):
        """An helper method for ``on*()`` handlers.

        It returns the number of mails which have been sent.
        """
        self._updateSubscriptionMapping(obj)
        users_by_label = self.getUsersToNotify(obj, what, extra_bindings)
        if self.isExtraSubscriptionsEnabled():
            users = users_by_label.get('', [])
            users.extend(self.getExtraSubscribersOf(self._getPath(obj)).items())
            users_by_label[''] = users

        n_sent = 0
        for label, users in users_by_label.items():
            users = self.removeUnAuthorizedSubscribers(users, obj)
            bindings = extra_bindings.copy()
            bindings['label'] = label
            for user, how in users:
                # Fetch the delivery utilities the user requested, then use
                # that to notify
                for h in how:
                    try:
                        delivery = getUtility(INotificationDelivery, h)
                    except:
                        # The method is not known, or the third party
                        # product that provided it was uninstalled
                        LOG.warning("Could not look up INotificationDelivery "\
                            "utility named '%s'", h)
                        continue
                    n_sent += delivery.notify(obj, user, what, label,
                        bindings)

        return n_sent
    #################################################################


    #################################################################
    ## Utility methods
    ###############################
    decPrivate('ignoreNotification')
    def ignoreNotification(self, obj):
        """Return whether notification have been set to be ignored for
        ``obj``.
        """
        ec = getExpressionContext(obj)
        for match_expr in self.getProperty('ignore_rules', ()):
            try:
                if self._match(match_expr, ec):
                    return True
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'ignore_rules' rule "\
                          "('%s') for '%s'",
                          match_expr, obj.absolute_url(1),
                          exc_info=True)
        return False


    decPrivate('getUsersToNotify')
    def getUsersToNotify(self, obj, what, ec_bindings=None):
        """Return a mapping from label to a list of user/how tuples,
        based on the passed ``what`` and ``ob``. ``what`` is one of the
        implemented notifications (*item_modification*, *wf_transition*,
        etc.). ``how`` is a string that says which delivery method to use.

        ``ec_bindings`` is a mapping which is injected into the
        expression context of the expression of the rules.
        """
        rules = self.getProperty('on_%s_users' % what, None)
        if rules is None:
            raise NotImplementedError, \
                "Notification on '%s' is not implemented." % what

        ec = getExpressionContext(obj, ec_bindings)
        users_by_label = {}
        ignore_next_rules = False
        for rule in rules:
            try:
                match_expr, users_expr = rule.split(RULE_DELIMITER, 1)
                parts = users_expr.split(RULE_DELIMITER)
                label = ''
                how = ('mail',)
                if len(parts) > 1:
                    users_expr, label = parts[:2]
                if len(parts) > 2:
                    how = tuple([p.strip() for p in parts[2:]])
            except ValueError:
                LOG.error("'%s' is not a valid rule "\
                          "('on_%s_users' on '%s')",
                          rule, what, obj.absolute_url(1))
                continue
            match_expr = match_expr.strip()
            users_expr = users_expr.strip()
            label = label.strip()
            users = users_by_label.get(label, [])
            try:
                if not self._match(match_expr, ec):
                    continue
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_users' rule "\
                          "('%s') for '%s'",
                          what, match_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            if users_expr == '*':
                users.extend([(u, how) for u in self.getAllUsers()])
                ignore_next_rules = True
            else:
                try:
                    users.extend([(u, how) for u in Expression(users_expr)(ec)])
                except ConflictError:
                    raise
                except:
                    LOG.error("Error in 'on_%s_users' rule "\
                              "('%s') for '%s'",
                              what, users_expr, obj.absolute_url(1),
                              exc_info=True)
            users_by_label[label] = users
            if ignore_next_rules:
                break
        return users_by_label


    decPrivate('getTemplate')
    def getTemplate(self, obj, what, ec_bindings=None):
        """Return the template to notify for the ``what`` of an object
        ``obj``, ``what`` being one of the implemented notification
        ("*item_modification*", "*wf_transition*", etc.), or ``None``
        if none could be found.

        ``ec_bindings`` is a mapping which is injected into the
        expression context of the expression of the rules.
        """
        rules = self.getProperty('on_%s_mail_template' % what, None)
        if rules is None:
            raise NotImplementedError, \
                'Notification on "%s" is not implemented.'

        ec = getExpressionContext(obj, ec_bindings)
        template = None
        for rule in rules:
            try:
                match_expr, template_expr = rule.split(RULE_DELIMITER)
                match_expr, template_expr = match_expr.strip(), template_expr.strip()
            except ValueError:
                LOG.error("'%s' is not a valid rule "\
                          "('on_%s_mail_template' on '%s')",
                          rule, what, obj.absolute_url(1))
                continue
            match_expr = match_expr.strip()
            template_expr = template_expr.strip()
            try:
                if not self._match(match_expr, ec):
                    continue
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_mail_template' rule "\
                          "('%s') for '%s'",
                          what, match_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            try:
                template = Expression(template_expr)(ec)
            except ConflictError:
                raise
            except:
                LOG.error("Error in 'on_%s_mail_template' rule "\
                          "('%s') for '%s'",
                          what, template_expr, obj.absolute_url(1),
                          exc_info=True)
                continue
            if type(template) == StringType:
                template = obj.restrictedTraverse(template, None)
            if template is not None:
                break
        return template


    decPrivate('getAllUsers')
    def getAllUsers(self):
        """Return a list of all user ids of the portal.

        **Warning:** this method may be costly if you rely on an
        external (non ZODB) user source. Use it at your own risk.
        """
        mtool = getToolByName(self, 'portal_membership')
        return mtool.listMemberIds()


    decPrivate('removeUnAuthorizedSubscribers')
    def removeUnAuthorizedSubscribers(self, subscribers, obj):
        """Return users from ``subscribers`` who are authorized to
        view ``obj``.
        """
        portal = getToolByName(self, 'portal_url').getPortalObject()
        mtool = getToolByName(self, 'portal_membership')
        filtered_subscribers = []
        for subscriber, notifymethod in subscribers:
            if self._anonymousShouldBeNotified(obj):
                filtered_subscribers.append((subscriber, notifymethod))
            else:
                ## We use '_huntUser()' and not 'mtool.getMemberById()'
                ## because the latter would provide a wrapped user
                ## object, with a specific context where the user is
                ## not allowed to view 'obj'.
                member = mtool._huntUser(str(subscriber), portal)
                if member is not None:
                    if member.has_permission('View', obj):
                        filtered_subscribers.append((subscriber, notifymethod))

        return filtered_subscribers


    def _match(self, expr, ec):
        """Return ``True`` if ``expr`` returns something which can be
        evaluated to ``True`` in the expression context (``ec``) or if
        ``expr`` is "*".
        """
        if expr == '*':
            return True
        expr = Expression(expr)
        return bool(expr(ec))


    def _getPath(self, obj):
        """Return path of ``obj``.

        A slash (``/``) is appended to the path if the object is
        folderish. The returned path is relative to the portal object.
        """
        utool = getToolByName(self, 'portal_url')
        path = utool.getRelativeContentURL(obj)
        path = '/' + path
        if not getattr(obj.aq_base, 'isPrincipiaFolderish', False):
            return path
        if path[-1] != '/':
            path += '/'
        return path


    def _getParents(self, path):
        """Get the parents of the item corresponding to ``path`` and
        return their respective path.

        Parents are returned from ``path`` to the portal root object.
        """
        if path == '/':
            return []
        if path[-1] == '/':
            path = path[:-1]
        parent = path[:path.rfind('/') + 1]
        parents = [parent]
        parents.extend(self._getParents(parent))
        return tuple(parents)


    def _getUID(self, obj):
        """Return UID of the object."""
        if not IATContentType.providedBy(obj) and not\
           IDexterityContent.providedBy(obj):
            return None

        portal_uidhandler = getToolByName(self, 'portal_uidhandler')
        uid = portal_uidhandler.queryUid(obj, None)
        if uid is None: ## Not yet registered
            uid = portal_uidhandler.register(obj)
        return uid


    def _anonymousShouldBeNotified(self, obj):
        """Return whether anonymous users should be notified, i.e.
        whether anonymous users can view ``obj``.
        """
        return 'Anonymous' in rolesForPermissionOn('View', obj)

    def _getChangeNote(self, obj):
        """ If obj has version control, get the last change note and
        return it, otherwise return None. """
        request_changenote = self.REQUEST.get('cmfeditions_version_comment')
        if request_changenote is not None:
            return request_changenote

        # CMFEditions might not be around, protect against that.
        try:
            archivist = getToolByName(self, 'portal_archivist')
        except AttributeError:
            return None

        try:
            history = archivist.getHistory(obj)
        except ArchivistUnregisteredError:
            return None

        if len(history)>0:
            return history[-1].sys_metadata['comment']
        return None

    #################################################################


    #################################################################
    ## Extra subscriptions settings
    ###############################
    decProtected('View', 'isExtraSubscriptionsEnabled')
    def isExtraSubscriptionsEnabled(self):
        """Return whether extra subscriptions are enabled."""
        return self.getProperty('extra_subscriptions_enabled')


    decProtected('View', 'isExtraSubscriptionsRecursive')
    def isExtraSubscriptionsRecursive(self):
        """Return whether extra subscriptions are recursive.

        Note that this method does not check whether extra
        subscriptions are enabled or not.
        """
        return self.getProperty('extra_subscriptions_recursive')
    #################################################################


    #################################################################
    ## Extra subscriptions logic
    ############################
    def _updateSubscriptionMapping(self, obj):
        """Update subscription mapping."""
        uid = self._getUID(obj)
        if not uid:
            return

        path = self._getPath(obj)
        known_path = self._uid_to_path.get(uid)
        if known_path != path:
            self._uid_to_path[uid] = path
            if known_path is not None:
                ## We have old informations for this object
                for key, value in self._subscriptions.items():
                    if key.startswith(known_path):
                        new_key = path + key[len(known_path) : ]
                        self._subscriptions[new_key] = value
                        del self._subscriptions[key]


    decPublic('currentUserHasSubscribePermission')
    def currentUserHasSubscribePermissionOn(self, obj):
        """Return whether the current user is allowed to subscribe to
        or unsubscribe from ``obj``.
        """
        return True

        if not IATContentType.providedBy(obj) and not \
                IDexterityContent.providedBy(obj) and not \
                IPloneSiteRoot.providedBy(obj):
            return False

        mtool = getToolByName(self, 'portal_membership')
        return mtool.checkPermission(SUBSCRIBE_PERMISSION, obj)


    decPublic('subscribeTo')
    def subscribeTo(self, obj, email=None, how=['mail']):
        """Subscribe ``email`` (or the current user if ``email`` is
        None) to ``obj``. You can pass the methods by which the user should
        be notified as a tuple using the ``how`` keyword argument.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if not self.currentUserHasSubscribePermissionOn(obj):
            raise Unauthorized
        elif email is not None:
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            ## FIXME: an anonymous user would like to subscribe
            ## his/her address. This has not yet been implemented, so
            ## we raise an exception.
            raise NotImplementedError
        else:
            self._updateSubscriptionMapping(obj)
            path = self._getPath(obj)
            subscribers = self._subscriptions.get(path, {})
            user = getSecurityManager().getUser().getId()
            subscribers[user] = tuple(how)
            self._subscriptions[path] = subscribers


    decPublic('unSubscribeFrom')
    def unSubscribeFrom(self, obj, email=None):
        """Unsubscribe ``email`` (or the current user if ``email`` is
        ``None``) from ``obj``.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if not self.currentUserHasSubscribePermissionOn(obj):
            raise Unauthorized
        elif email is not None:
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            ## FIXME: an anonymous user would like to unsubscribe
            ## his/her address. This has not yet been implemented, so
            ## we raise an exception.
            raise NotImplementedError
        else:
            self._updateSubscriptionMapping(obj)
            path = self._getPath(obj)
            subscribers = self._subscriptions.get(path, {})
            user = getSecurityManager().getUser().getId()
            try:
                del subscribers[user]
                self._subscriptions[path] = subscribers
            except KeyError:
                pass ## User was not subscribed.


    decPublic('unSubscribeFromObjectAbove')
    def unSubscribeFromObjectAbove(self, obj, email=None):
        """Find folderish items above ``obj`` and unsubscribe
        ``email`` (or the current user if ``email`` is ``None``) from
        the first one (s)he is subscribed to.

        If ``user`` is subscribed to ``obj``, this method is
        equivalent to ``unSubscribeFrom(obj, user)``.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if not self.currentUserHasSubscribePermissionOn(obj):
            raise Unauthorized
        elif email is not None:
            if not EMAIL_REGEXP.match(email):
                raise InvalidEmailAddress
            ## FIXME: an anonymous user would like to unsubscribe
            ## his/her address. This has not yet been implemented, so
            ## we raise an exception.
            raise NotImplementedError
        else:
            self._updateSubscriptionMapping(obj)
            utool = getToolByName(obj, 'portal_url')
            portal = utool.getPortalObject()
            portal_container = portal.aq_inner.aq_parent
            while obj != portal_container:
                if self.isSubscribedTo(obj, as_if_not_recursive=True):
                    self.unSubscribeFrom(obj)
                    break
                obj = obj.aq_parent


    decPublic('isSubscribedTo')
    def isSubscribedTo(self, obj, email=None,
                       as_if_not_recursive=False):
        """Return whether ``email`` (or the current user if ``email``
        is ``None``) is subscribed to ``obj``.

        If ``as_if_not_recursive`` is ``True``, this method acts as if
        the recursive mode was off.
        """
        if not self.isExtraSubscriptionsEnabled():
            raise DisabledFeature

        if email is None:
            ## Yes, 'email' is actually the id of the current user.
            email = getSecurityManager().getUser().getId()
        self._updateSubscriptionMapping(obj)
        path = self._getPath(obj)
        subscribers = self.getExtraSubscribersOf(path,
                                                 as_if_not_recursive)
        return subscribers.has_key(email)


    decPrivate('getExtraSubscribersOf')
    def getExtraSubscribersOf(self, path, as_if_not_recursive=False):
        """Return users or email addresses which are subscribed to
        the given path.

        This method returns a mapping of users to tuples of notification
        methods, that is, each user can subscribe to be notified in more
        than one way.  If ``as_if_not_recursive`` is ``True``, this
        method acts as if the recursive mode was off.
        """
        subscribers = self._subscriptions.get(path, {}).copy()
        if self.isExtraSubscriptionsRecursive() and \
                not as_if_not_recursive:
            if path[-1] == '/':
                path = path[:-1]
            i = path.rfind('/')
            if i != -1:
                parent = path[:i + 1]
                subscribers.update(self.getExtraSubscribersOf(parent))
        return subscribers
    #################################################################


InitializeClass(NotificationTool)



"""Test pure Python functions of CMFNotification ``utils`` module.

$Id$
"""

from unittest import TestCase

from Products.CMFNotification.utils import encodeMailHeaders


MSG1 = '''\
From: george@example.com
Sender: George <george@example.com>
To: eric@example.com
Subject: My guitar gently weeps

Eric, my guitar weeps. Gently, sure, but it weeps. What shall I do?
'''

MSG1_ENCODED = '''\
From: george@example.com
Sender: =?utf-8?q?George?= <george@example.com>
To: eric@example.com
Subject: =?utf-8?q?My_guitar_gently_weeps?=

Eric, my guitar weeps. Gently, sure, but it weeps. What shall I do?
'''

MSG2 = '''\
From: Andrew <andrew@example.com>, Dick <dick@example.com>
Sender: Andrew <andrew@example.com>
Reply-To: Andrew <andrew@example.com>
CC: Jim <jim@example.com>
Subject: What happened to the door?

What happened to the door?! It was red, and now it's black. This is
not funny. And Jim, get off that snake, immediately!
'''

MSG2_ENCODED = '''\
From: =?utf-8?q?Andrew?= <andrew@example.com>, =?utf-8?q?Dick?= <dick@example.com>
Sender: =?utf-8?q?Andrew?= <andrew@example.com>
Reply-To: =?utf-8?q?Andrew?= <andrew@example.com>
CC: =?utf-8?q?Jim?= <jim@example.com>
Subject: =?utf-8?q?What_happened_to_the_door=3F?=

What happened to the door?! It was red, and now it's black. This is
not funny. And Jim, get off that snake, immediately!
'''

MESSAGES = ((MSG1, MSG1_ENCODED),
            (MSG2, MSG2_ENCODED),
            )


class TestEncodeMailHeaders(TestCase):
    """Various tests for ``utils.encodeMailHeaders()``."""

    def testEncoding(self):
        for raw, expected in MESSAGES:
            encoded = encodeMailHeaders(raw, 'utf-8')
            self.failUnlessEqual(encoded, expected)

def test_suite():
    import unittest
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestEncodeMailHeaders))
    return suite
